/*******************************************************************************
 $Author: dmc $ 
 $Revision: 2898 $
 $Date: 2021-03-19 16:08:01 -0300 (Fri, 19 Mar 2021) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/calc_shot.js $
*******************************************************************************/

// TM600X and TM-600G2 clock speed, in counts per millisecond
var G1clock = 125000; // 125 MHz
var G2clock = 32000 / 2; // 32 MHz xtal divided by two

var _performance = require('./lib/utils/performance.js');
var fullborejs = require("./fullbore"); // only until it can be fully replaced
eval(require("fs").readFileSync("./smt.common.lib/public/javascripts/math.js")+"");

var mic_mapping = require('./mic_mapping.js');


function calc_mic_mapping(shot) {
	// determine mic mapping - shift timings to different mic positions
	// use default, or global setting, or frame overrides
	shot.res.mic_mapping = mic_mapping.get_mic_mapping(shot);
	//var logical_mics_in_use = mic_mapping.logical_mics_in_use(mic_mapping.get_mic_mapping(shot)); // used below...
	var mic = shot.res.mic = mic_mapping.phys_to_logical(shot.res.mic_mapping,shot.shot.counters);
	var test=false; if (test) {
		var phys=mic_mapping.logical_to_phys(res.mic_mapping,mic);
		console.log("DMC 0",shot.shot.counters)
		console.log("DMC 1",mic)
		console.log("DMC 2",phys)
	}
}


module.exports.solve_impact = function(shot) { // only valid for geom == 'IMPACT'
	// must set shot.res.{xy_valid,v=0,x=undef,y=undef,theta,phi,impact_mics=[..]}
	calc_mic_mapping(shot);

	shot.res.impact_mics = shot.res.mic || [];
	shot.res.xy_valid = true;
	shot.res.v = 555; // unfortuntely 0 means invalid
	shot.res.x = 0;
	shot.res.y = 0;
	shot.res.theta = 0;
	shot.res.phi = 0;

}


module.exports.solve_iterative = function(shot) { // only valid for geom == 'STD'
	var start = _performance();

	// calc_initial has been done by caller; res.(device_type, clock, temp, soundspeed) already set up.

	// output directly to shot itself
	if (!shot.res) shot.res = {};
	var res = shot.res;

	var spacing = shot.geom.chrono_spacing;
	var mic_w = shot.geom.mic_width;
	var mic_h = shot.geom.mic_height;

	calc_mic_mapping(shot);
	var mic = res.mic;

	res.theta = 0;
	res.phi = 0;
	
	//console.log(mic);
	
	var mic_rot = mic.slice(); // rotate-able timings

	// mic positions, needed for rotating
	var mic_pos = mic_mapping.get_mic_pos(shot); // fetch canonical mic positions

	var iter_n = 5; // seems best
	// console.log('solve_iterative - start iter (' + _performance(start) + 'ms)');
	for (res.iter = 1; res.iter <= iter_n; res.iter++) {
		var start0 = _performance();
		
		//console.log(mic);
		
		// v, v13/v26/v47/v58 are set by calc_velocity only if "valid" (meaning, supersonic and <10,000m/s)
		// v = 0 for invalid velocity
		calc_velocity(res, mic_rot, spacing, res.clock, res.soundspeed);
		if (!res.v) break; // no valid velocities
				
		// for each bad chrono, disable both mics so they cannot skew the result
		var mic_rot_valid = mic_rot.slice();
		var count_errors = 0;
		if (mic_rot[1-1] >= 0 && mic_rot[3-1] >= 0 && !res.v13) { mic_rot_valid[1-1] = mic_rot_valid[3-1] = -1; count_errors++; }
		if (mic_rot[2-1] >= 0 && mic_rot[6-1] >= 0 && !res.v26) { mic_rot_valid[2-1] = mic_rot_valid[6-1] = -1; count_errors++; }
		if (mic_rot[4-1] >= 0 && mic_rot[7-1] >= 0 && !res.v47) { mic_rot_valid[4-1] = mic_rot_valid[7-1] = -1; count_errors++; }
		if (mic_rot[5-1] >= 0 && mic_rot[8-1] >= 0 && !res.v58) { mic_rot_valid[5-1] = mic_rot_valid[8-1] = -1; count_errors++; }

if(0){
		if (res.iter > 1 && count_errors > 0 || count_errors > 1) { // DMC 23-Nov-2020, a single mis-fired mic will
			// corrupt a velocity pair.  If subsequent rotations make that subsonic we'll get a ghost shot (Cleveland 2020)
			res.v = 0; // this will mark shot as invalid
			res.calc_shot_err_msg = "invalidate as likely ghost shot due to count_errors="+count_errors;
			break;
		}
} else {
		// returning res.v_error < 0 means v_error is a count of how many chrono pairs had bad readings...
		// ... meaning that this shot is probably a ghost shot

}		
		// res.x/y
		// res.xy_valid = true if valid
		calc_position(res, mic_rot_valid, res.v, res.clock, res.soundspeed, mic_w, mic_h);
		if (!res.xy_valid) break; // no position solution
		
//		console.log("x = " + res.x + ", y = " + res.y);
//console.log("DMC calc_shot",data.config.enable_angle_cal);

		//console.log("step " + res.iter, res);

		// done
		if (res.iter == iter_n) break;
		
		// simple method only for shots way off target
		//if (abs(res.x) > mic_w || abs(res.y) > mic_h) break;
		
		// theta, phi
		// do not try to exit early if apparently only a small rotation is needed
		// when xy changes, the rotation takes place around a different origin
		// we may need more iterations and it is difficult to tell
		//
		// DMC 9-Feb-2017 change, for Bisley:
		// -- use the "angle search" method for all the iterations but the last
		// -- on the last iteration, we calculate using fixed theta/phi
		var bisley_behaviour = true;
		var bisley_behaviour = false; // we need to find a GUI way to turn this on/of, or make a decision on this.
//		if ('undefined' != typeof data.config.bisley_behaviour) bisley_behaviour = data.config.bisley_behaviour;
		if ('undefined' != typeof shot.geom.bisley_behaviour) bisley_behaviour = shot.geom.bisley_behaviour;
//console.log("*** DMC BBB ***",shot.geom.bisley_behaviour);
		
		if (!bisley_behaviour && data.config.enable_angle_cal) { // fixed angle set by user
			
			// no rotation to apply, can definitely exit
			if (!shot.geom.theta_cal && !shot.geom.phi_cal) {
				//console.log("exiting, fixed angles are zero");
				break;
			}
			
			// apply, mils to radians
			if (shot.geom.theta_cal) res.theta = shot.geom.theta_cal * 0.001; // mils to radians
			if (shot.geom.phi_cal) res.phi = shot.geom.phi_cal * 0.001;
			
		} else if (bisley_behaviour && data.config.enable_angle_cal && res.iter == iter_n-1) { // on the last iteration
			res.bisley_theta = res.theta;
			res.bisley_phi = res.phi;
			res.bisley_behaviour = true;

			// apply, mils to radians
			if (shot.geom.theta_cal) res.theta = shot.geom.theta_cal * 0.001; // mils to radians
			if (shot.geom.phi_cal) res.phi = shot.geom.phi_cal * 0.001;
			
		} else { // measure angle using multiple chronos

			//console.log("checking angle");
		
			var res_delta = {};
			calc_thetaphi(res_delta, res.x, res.y, res.v13, res.v26, res.v47, res.v58, res.soundspeed, mic_w, mic_h);
			if (!res_delta.valid) break; // no angle change can be calculated: exit
			// console.log('solve_iterative - iter#',res.iter,'dtheta=', 1000*res_delta.theta, "dphi=",1000*res_delta.phi)
			res.theta += res_delta.theta;
			res.phi += res_delta.phi;
			
		}
		
		//console.log("rotating by " + res.theta + ", " + res.phi);
		
		// adjust mic timings to rotate frame
		var min_dt = 0;
		for (i in mic) {
			
			if (mic[i] < 0) continue;
			
			// new mic position rotated by theta/phi around x/y
			var v_new = xyz_rotate(mic_pos[i], res.phi, -res.theta, 0, {x: res.x, y: res.y, z: 0});
			
			// dt = difference between simulated timings for both mic positions
			var t_old = calc_timing(mic_pos[i], res.x, res.y, res.v, res.theta, res.phi, res.clock, res.soundspeed);
			var t_new = calc_timing(v_new, res.x, res.y, res.v, res.theta, res.phi, res.clock, res.soundspeed);
			var dt = t_new - t_old;
			
			mic_rot[i] = mic[i] + dt;
			
			if (mic_rot[i] < min_dt) min_dt = dt;

			//console.log("mic " + i + " from " + mic[i] + " to " + mic_rot[i]);
		}
		
		// re-zero valid timings to prevent negative
		if (min_dt) for (i in mic) if (mic[i] >= 0) mic_rot[i] -= min_dt;
		// console.log('solve_iterative - iter#',res.iter,' (' + _performance(start0) + 'ms)');
	
	}
	
	res.x += shot.geom.x_cal;
	res.y += shot.geom.y_cal;

	//console.log("calc_shot done!\n");
	//console.log(shot);
	// console.log('solve_iterative - done (' + _performance(start) + 'ms)');
	// shot.is_starred is now done by calc_scorezone
	// shot.is_starred = false;
//	if (data.config.star_shot_missing_mic) shot.is_starred = shot.is_starred || shot.whack_raw_v; // at least on
//	shot.is_starred = shot.whack_raw_v || shot.whack_v || shot.whack_xy_err || shot.whack_v_err; // now done in recv_shot()

}

// simulate a shot (new method, with 8 mics and angles)
// just compute timings from given shot params
module.exports.simulate_timings = function(shot) {

	var mic_timings = []; // logical channels
	var spacing = shot.sim_data.geom.chrono_spacing;
	var mic_w = shot.sim_data.geom.mic_width;
	var mic_h = shot.sim_data.geom.mic_height;
	
	var x = shot.sim_data.x - shot.sim_data.geom.x_cal;
	var y = shot.sim_data.y - shot.sim_data.geom.y_cal;
	var v = shot.sim_data.v;
	var theta = shot.sim_data.theta;
	var phi = shot.sim_data.phi;
	var temp = shot.shot.G2_temp_rdg = shot.sim_data.temp;

	var clock = G2clock;
	var soundspeed = 331.5 * sqrt(1+(temp/273.15));

	// mic positions
	var mic_pos = mic_mapping.get_mic_pos(shot); // fetch canonical mic positions

	for (var i = 0; i < mic_pos.length; i++) {
		mic_timings[i] = round(calc_timing(mic_pos[i], x, y, v, theta, phi, clock, soundspeed));
	}

	// we now have the (absolute, t0-ref'd) timings for canonically-located mics in mic_timings[]
	// produce reverse-mapped physical timer readings in shot.shot.counters[] (normalized to t0)

	shot.shot.counters = mic_mapping.logical_to_phys(mic_mapping.get_mic_mapping(shot),mic_timings);

	var t0 = -1;
	for (i in shot.shot.counters) { // find t0
		if (shot.shot.counters[i] == -1) continue;
		if (t0 != -1 &&  shot.shot.counters[i] > t0) continue;
		t0 = shot.shot.counters[i];
	}

	for (i in shot.shot.counters) if (shot.shot.counters[i] != -1) shot.shot.counters[i] -= t0;
	//shot.shot.raw_counters = shot.shot.counters.slice();
	//shot.raw_t0 = t0; // DEBUG
}



// calc_ functions (res, inputs...)
// output is put into res, not the shot itself
// input is given as individual parts if possible
// inputs are not modified


// set device_type, clock, temp, soundspeed
// set temp_invalid only if true (fallback warning)
module.exports.calc_initial = function calc_initial(res, shot) {
	
	if (shot.source == "sim_sw" || shot.source == "SOLO" || shot.source == "G2" || shot.connection_type == "TCP") {
		res.device_type = "G2";
		res.clock = G2clock;
	} else {
		res.device_type = "G1";
		res.clock = G1clock;
	}
	
//	console.log("DMC: calc_initial shot=",shot);
	var G1_temp_counts = shot && shot.shot && shot.shot.temp_rdg;
    var G2_temp = shot && shot.shot && shot.shot.G2_temp_rdg; // may be undefined
    var IP_temp = global.ip_temperature;
    
    // IP temperature in use
    if (IP_temp > -100) {
    	res.temp = IP_temp + shot.geom.temp_cal;
	
	// working G2
    } else if ('undefined' != typeof G2_temp && G2_temp > -25 && G2_temp < 70) {
		res.temp = G2_temp + shot.geom.temp_cal;
        fullborejs.globalTemperature.addValue(res.temp);
    	
	// working G1
    } else if (G1_temp_counts && G1_temp_counts <= 1020 && G1_temp_counts >= 10) {
    	res.temp = fullborejs.thermistor_lookup_table[G1_temp_counts] + shot.geom.temp_cal;
        fullborejs.globalTemperature.addValue(res.temp);
    	
	// invalid, but apply temp_cal so it can be modified
    } else {
    	res.temp_invalid = true;
    	res.temp = 20 + shot.geom.temp_cal; // use 20C(tweakable by user) if we have nothing better...
    	// res.temp = 10 + shot.geom.temp_cal; // DMC Feb-2017: seasonally-appropriate-Bisley value = 10C ;-)
		// But, if we are an S3/S10 and we have other devices in the system that are reporting valid temperatures,
        // we can use that global moving average temperature reading.  This is used e.g. at Kasseedorf.
        
        // ADAM disabled this in Lodi because it is screwing things up when the temp sensor isnt plugged in
        
        //if (fullborejs.globalTemperature.hasValues()) {
    	//	res.temp = fullborejs.globalTemperature.average;
   		//}
    }
    res.temp = trim_num(res.temp,2);
    res.soundspeed = trim_num(331.5 * sqrt(1+(res.temp/273.15)), 2);
}

// set res.v
// res.v = 0 means shot is invalid -- this is what is expected by server_frames.js and common.server.js
// NEW: res.v_error < 0 if one or more chrono readings invalid (subsonic or >10,000m/s)
// for each chrono, if valid, set v13,v26,v47,v58
function calc_velocity(res, mic, spacing, clock, soundspeed) {

	// calculate velocity on each chrono and only store if valid
	var count = 0;
	res.v = 0;
	res.v_error = 0;
	delete res.v13;
	delete res.v26;
	delete res.v47;
	delete res.v58;
	
	if (mic[1-1] >= 0 && mic[3-1] >= 0) {
		var v13 = clock * spacing / (mic[1-1] - mic[3-1]);
		if (!res.raw_v13) res.raw_v13 = v13;
		if (v13 > soundspeed && v13 < 10000) {
			res.v13 = v13;
			res.v += v13;
			count++;
		} else {
			res.v_error--;
		}
	}

	if (mic[2-1] >= 0 && mic[6-1] >= 0) {
		var v26 = clock * spacing / (mic[2-1] - mic[6-1]);
		if (!res.raw_v26) res.raw_v26 = v26;
		if (v26 > soundspeed && v26 < 10000) {
			res.v26 = v26;
			res.v += v26;
			count++;
		} else {
			res.v_error--;
		}
	}

	if (mic[4-1] >= 0 && mic[7-1] >= 0) {
		var v47 = clock * spacing / (mic[4-1] - mic[7-1]);
		if (!res.raw_v47) res.raw_v47 = v47;
		if (v47 > soundspeed && v47 < 10000) {
			res.v47 = v47;
			res.v += v47;
			count++;
		} else {
			res.v_error--;
		}
	}

	if (mic[5-1] >= 0 && mic[8-1] >= 0) {
		var v58 = clock * spacing / (mic[5-1] - mic[8-1]);
		if (!res.raw_v58) res.raw_v58 = v58;
		if (v58 > soundspeed && v58 < 10000) {
			res.v58 = v58;
			res.v += v58;
			count++;
		} else {
			res.v_error--;
		}
	}
	
	// average velocity
	if (count > 0) {
		res.v /= count;
		if (res.v_error == 0) {
			var mad = 0; // Mean Absolute Deviation
			if (res.v13) mad += abs(res.v13 - res.v);
			if (res.v26) mad += abs(res.v26 - res.v);
			if (res.v47) mad += abs(res.v47 - res.v);
			if (res.v58) mad += abs(res.v58 - res.v);
			res.v_error = trim_num(mad/count,2);
		}
	}
}

// set x, y
// requires at least three of mic 1,2,4,5 to be valid
// set xy_valid = true or false
// set xy_error only if all four mics are used
function calc_position(res, mic, v, clock, soundspeed, mic_w, mic_h) {
	
	//console.log("calc_position");
	
	var ac = soundspeed / clock;
	var vc = v / clock;
	var k2inv = (vc*vc - ac*ac) / (vc*vc * ac*ac);
	
	var sol125 = null;
	var sol214 = null;
	var sol541 = null;
	var sol452 = null;

	var count = 0;
	res.x = 0;
	res.y = 0;

	if (mic[1-1] >= 0 && mic[2-1] >= 0 && mic[5-1] >= 0) {
		var fullbore_res = {};
		if (fullborejs.calc_fullbore(fullbore_res, mic_h, mic_w, 0, 0, k2inv, mic[1-1], mic[2-1], mic[5-1])) {
			sol125 = {x: fullbore_res.x0 - mic_w/2, y: fullbore_res.y0 - mic_h/2};
			res.x += sol125.x;
			res.y += sol125.y;
			count++;
			//console.log("125", sol125);
		}
	}

	if (mic[2-1] >= 0 && mic[1-1] >= 0 && mic[4-1] >= 0) {
		var fullbore_res = {};
		if (fullborejs.calc_fullbore(fullbore_res, -mic_h, mic_w, 0, mic_h, k2inv, mic[2-1], mic[1-1], mic[4-1])) {
			sol214 = {x: fullbore_res.x0 - mic_w/2, y: fullbore_res.y0 - mic_h/2};
			res.x += sol214.x;
			res.y += sol214.y;
			count++;
			//console.log("214", sol214);
		}
	}

	if (mic[5-1] >= 0 && mic[4-1] >= 0 && mic[1-1] >= 0) {
		var fullbore_res = {};
		if (fullborejs.calc_fullbore(fullbore_res, mic_h, -mic_w, mic_w, 0, k2inv, mic[5-1], mic[4-1], mic[1-1])) {
			sol541 = {x: fullbore_res.x0 - mic_w/2, y: fullbore_res.y0 - mic_h/2};
			res.x += sol541.x;
			res.y += sol541.y;
			//console.log("541", sol541);
			count++;
		}
	}

	if (mic[4-1] >= 0 && mic[5-1] >= 0 && mic[2-1] >= 0) {
		var fullbore_res = {};
		if (fullborejs.calc_fullbore(fullbore_res, -mic_h, -mic_w, mic_w, mic_h, k2inv, mic[4-1], mic[5-1], mic[2-1])) {
			sol452 = {x: fullbore_res.x0 - mic_w/2, y: fullbore_res.y0 - mic_h/2};
			res.x += sol452.x;
			res.y += sol452.y;
			count++;
			//console.log("452", sol452);
		}
	}
	
	if (count > 0) {
		res.x /= count;
		res.y /= count;
		res.xy_valid = true;
	} else {
		res.xy_valid = false;
	}
	
	if (count == 4) {
		var d125 = sqrt((sol125.x - res.x) * (sol125.x - res.x) + (sol125.y - res.y) * (sol125.y - res.y));
		var d214 = sqrt((sol214.x - res.x) * (sol214.x - res.x) + (sol214.y - res.y) * (sol214.y - res.y));
		var d541 = sqrt((sol541.x - res.x) * (sol541.x - res.x) + (sol541.y - res.y) * (sol541.y - res.y));
		var d452 = sqrt((sol452.x - res.x) * (sol452.x - res.x) + (sol452.y - res.y) * (sol452.y - res.y));
		res.xy_error = (d125 + d214 + d541 + d452) / 4;
		
		//console.log("mean_dev", res.xy_error);
		
	} else {
		res.xy_error = 0;
	}
	
}

// set theta, phi
// computes rotation around xy that is needed to equalize the four velocities
function calc_thetaphi(res, x, y, v13, v26, v47, v58, soundspeed, mic_w, mic_h) {
	
	//console.log("checking theta/phi");
	
	res.theta = res.phi = 0;

	// mach numbers
	var m13, m26, m47, m58;
	if (v13) m13 = sqrt((v13/soundspeed)*(v13/soundspeed) - 1);
	if (v26) m26 = sqrt((v26/soundspeed)*(v26/soundspeed) - 1);
	if (v47) m47 = sqrt((v47/soundspeed)*(v47/soundspeed) - 1);
	if (v58) m58 = sqrt((v58/soundspeed)*(v58/soundspeed) - 1);

	var sol = {};

	// only use gamma corrections on target
	var range = min(mic_w, mic_h) / 2;
	if (x*x + y*y < range*range) {

		// angles from xy to corners
		var d1 = sqrt((x - -mic_w/2)*(x - -mic_w/2) + (y - -mic_h/2)*(y - -mic_h/2));
		var d2 = sqrt((x - -mic_w/2)*(x - -mic_w/2) + (y - mic_h/2)*(y - mic_h/2));
		var d4 = sqrt((x - mic_w/2)*(x - mic_w/2) + (y - mic_h/2)*(y - mic_h/2));
		var d5 = sqrt((x - mic_w/2)*(x - mic_w/2) + (y - -mic_h/2)*(y - -mic_h/2));
		var phi1_g1 = d1 / (y - -mic_h/2);
		var phi1_g2 = -d2 / (y - mic_h/2);
		var phi2_g1 = d5 / (y - -mic_h/2);
		var phi2_g2 = -d4 / (y - mic_h/2);
		var theta1_g1 = d1 / (x - -mic_w/2);
		var theta1_g2 = -d5 / (x - mic_w/2);
		var theta2_g1 = d2 / (x - -mic_w/2);
		var theta2_g2 = -d4 / (x - mic_w/2);

		// solutions for theta/phi from each chrono pair
		if (v13 && v58) sol.theta1 = -atan(theta1_g1 * theta1_g2 * (v13 - v58) / (theta1_g1 * v13 * m58 + theta1_g2 * m13 * v58));
		if (v26 && v47) sol.theta2 = -atan(theta2_g1 * theta2_g2 * (v26 - v47) / (theta2_g1 * v26 * m47 + theta2_g2 * m26 * v47));
		if (v13 && v26) sol.phi1 = -atan(phi1_g1 * phi1_g2 * (v13 - v26) / (phi1_g1 * v13 * m26 + phi1_g2 * m13 * v26));
		if (v58 && v47) sol.phi2 = -atan(phi2_g1 * phi2_g2 * (v58 - v47) / (phi2_g1 * v58 * m47 + phi2_g2 * m58 * v47));

	// simpler method, gamma = 1
	// more reliable off target
	} else {
		
		// solutions for theta/phi from each chrono pair
		if (v13 && v58) sol.theta1 = -atan((v13 - v58) / (v13 * m58 + m13 * v58));
		if (v26 && v47) sol.theta2 = -atan((v26 - v47) / (v26 * m47 + m26 * v47));
		if (v13 && v26) sol.phi1 = -atan((v13 - v26) / (v13 * m26 + m13 * v26));
		if (v58 && v47) sol.phi2 = -atan((v58 - v47) / (v58 * m47 + m58 * v47));
	}
	
	//console.log(sol);
	
	var count = 2;
	res.valid = true;
	
	// apply valid solutions
	if (v13 && v58 && v26 && v47) res.theta = (sol.theta1 + sol.theta2) / 2;
	else if (v13 && v58) res.theta = sol.theta1;
	else if (v26 && v47) res.theta = sol.theta2;
	else count--;
	
	if (v13 && v26 && v58 && v47) res.phi = (sol.phi1 + sol.phi2) / 2;
	else if (v13 && v26) res.phi = sol.phi1;
	else if (v58 && v47) res.phi = sol.phi2;
	else count--;
	
	if (!count) res.valid = false; // no changes made
	
}

// simulate one mic timing for a shot
function calc_timing(mic_pos, x, y, v, theta, phi, clock, soundspeed) {

	var ac = soundspeed / clock;
	var vc = v / clock;
	var k = sqrt((vc*vc * ac*ac) / (vc*vc - ac*ac));

	// unit direction of bullet from target towards source
	var v_path = {x: sin(theta) * cos(phi), y: sin(phi), z: -cos(theta) * cos(phi)};

	// vector from bullet impact to mic
	var v_bullet_to_mic = {x: mic_pos.x - x, y: mic_pos.y - y, z: mic_pos.z};

	// project along bullet path to find distance to shockwave propagation start point
	var path_len = v_bullet_to_mic.x * v_path.x + v_bullet_to_mic.y * v_path.y + v_bullet_to_mic.z * v_path.z;
	var v_shock_start = {x: x + v_path.x * path_len, y: y + v_path.y * path_len, z: v_path.z * path_len};
	var shock_len = xyz_dist(v_shock_start, mic_pos);

	return shock_len / k + path_len / vc;
}


// we want to aggressively trim the precision of the numbers delivered
// in a shot string.
function trim_num(nn,ndigs) {
		if (typeof ndigs == 'undefined' || ndigs < 0 || ndigs > 5) ndigs = 0;
		if (typeof nn == 'number') return +nn.toFixed(ndigs); // '+' needed to return 'number' not 'string'
		return nn;
}

