/*******************************************************************************
 common.server.js: server functionalities common across SMT product lines

 $Author: dmc $
 $Revision: 3226 $
 $Date: 2024-01-19 15:19:31 -0400 (Fri, 19 Jan 2024) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/common.server.js $
 $Id: common.server.js 3226 2024-01-19 19:19:31Z dmc $

 late-mic:
 one mic misfire - screen by diagonal angle deviations from mean
 two mics: if bullet speed < 370(say) and subsonic chrono, screen out (maybe also at plunge angle anomaly?)

*******************************************************************************/

// temp'y workarounds while making this a require() module:
var log = console.log;
var log = function() {};



var EventEmitter = require( "events" ).EventEmitter;
global.qndEventEmitter = new EventEmitter();

//var mic_mapping = require('./smt.common.lib/mic_mapping.js');
var mic_mapping = require('./mic_mapping.js');
var _performance = require('./lib/utils/performance.js');
var fs = require("fs"); // file system
eval(fs.readFileSync("./smt.common.lib/public/javascripts/utils.js")+""); // common utils

var string_stats_timers = {}; // for more efficient calling of string stats

// general shot handler: add to string
// currently all measurements (valid or not) are saved in string
module.exports.recv_shot = function recv_shot(shot, frame_id, data) {
	// debugging:
	var dbglog = console.log;
	// no debug:
	if (!data.config.option_debuginfo) dbglog = function() {};
	
	// defensive:
	if (!shot.res) shot.res = {};

	//log("recv_shot:",shot);

	var frame = data.frames[frame_id];
	var str = data.strings[frame.string_id];
	var face = data.faces[str.face_id];
	var device = data.devices[frame.assoc_device_id];
	
	shot.ping_health = device && device.ping_health; // capture SOLO health stats, if available
	
	// apply shot_params from string
	if (str.shot_params) update_object(shot, str.shot_params); // copy scope settings etc (plotomatic mode)

	// add parallax correction if present in targetface
	// CAREFUL!!!!
	shot.rawX = shot.x;
	shot.rawY = shot.y;
    
	if (typeof(shot.x) == "number" && typeof(shot.y) == "number") {
		if (face.parallax_offset_x) { shot.x = face.parallax_offset_x + shot.rawX; }
		if (face.parallax_offset_y) { shot.y = face.parallax_offset_y + shot.rawY; }
	}

	// calc_shot_valid() behaviour:
	// if learning - will accept all; then call recalc_.. (might transition)
	// if learned - may mark as invalid
	//.. recalc_string_scores - re-decide learned/learning, re-mark valid/invalid, 

	// valid?  sets shot.valid
	// reject? sets shot.reject

	calc_shot_valid(shot,str);
	calc_shot_reject(shot,str);

	if (shot.reject_shot) {
console.log("recv_shot: reject");
		return; // no further processing		
	}

	recv_shot_assign_shooter(shot,str); // does calc_scorezone

	var crud = {};
	crud.strings = {};
	crud.strings[str._id] = {};

	if ( shot.nmics_fired === 1 ) {
		if ( str.taps === undefined ) {
			str.taps = [];
			str.tapsCount = 1;
		} else {
			str.tapsCount++;
		}
		str.taps.push( { ts: shot.ts, mics: shot.mics } );
		if ( str.taps.length > 10 ) {
			str.taps.shift( );
		}
		crud.strings[str._id].taps = str.taps;
		crud.strings[str._id].tapsCount = str.tapsCount;
	}
	// DMC: report a "tap" as a "shot".  This is NEEDED for now, for example in order to get/change temperature
	// And besides, I think the cost/overhead is overstated.
	crud.strings[str._id].shots = [];
	crud.strings[str._id].shots[str.shots.length] = shot;
	//
	if (shot.valid) {
		str.validshots = str.validshots || [];
		crud.strings[str._id].validshots = [];
		crud.strings[str._id].validshots[str.validshots.length] = shot;
	}

	/*
			update_string_stats(str._id); // update_string_stats() is O(N^2) and quite expensive;
			// 700ms on G2 for 100 shots
			// console.log('update string stats - (' + _performance(start) + 'ms)');
		}, 750); // wait 0.75s after last shot fired before calc'ing stats 
		// SOLO 250 vs. 750:
		// 1250, 2 clients, 100 shots in 45s
		// 750, 2 clients, 100 shots in 46s
		// 250, 2 clients, 100 shots in 52s
	*/
	var update_string_stats_delay = 750;

	
	apply_crud(crud);
	// SOLO:
	// 2 clients: if we don't send the shots, we can do 100 shots in 36s (vs 48s when sending)
	// 10 clients: if we don't send the shots, we can do 100 shots in 40s (vs 80s when sending)
	
	// TODO: auto-learn acceptable incoming angles, for each string.
	//		str.autolearn_angles = true;

	
	// auto update string stats (later)
	if (shot.valid) {
		clearTimeout(string_stats_timers[str._id]);
		if (str.autolearn_angles) {
			global.update_string_stats(str._id); // immediate call
		} else {
// console.log("DBG clearTimeout string_stats_timers",str._id);
// 			clearTimeout(string_stats_timers[str._id]);
			string_stats_timers[str._id] = setTimeout(function () {
				var start = _performance();
				global.update_string_stats(str._id); // update_string_stats() is O(N^2) and quite expensive;
			}, update_string_stats_delay);
		}
	}

	// send valid shot snippet to other G2s using UDP broadcast
	if (shot.valid) {
		recent_shot_valid = true;
		recent_shot_x = shot.x;
		recent_shot_y = shot.y;
	}
	
}

var angle_outside = function (a_check, a_nom, a_tol) {
	if (typeof a_check == 'undefined') return false;
	return a_tol && (a_check < a_nom - a_tol || a_check > a_nom + a_tol);
}

var calc_shot_valid = function (shot_or_index,str) {
	var shot_index = shot_or_index, shot;
	if ('number' == typeof shot_index) {
		shot = str.shots[shot_or_index];
	} else {
		shot = shot_or_index;
		shot_index = -1; // not in str[]
	}
	// console.log("DMC calc_shot_valid",shot_index,typeof shot);
	if ('object' != typeof shot) return;
	// defensive:
	if (!shot.res) shot.res = {};

	var face = data.faces[str.face_id];
	// console.log("DMC  calc_shot_valid",shot_or_index);
	// debugging:
	var dbglog = console.log;
	// no debug:
	if (!data.config.option_debuginfo) dbglog = function() {};

	if (shot.res.method == "IMPACT") {
		shot.valid = shot.res.impact_mics.length != 0; // valid iff a mic fired
	}

	// optimization: short-circuit this routine if possible; saves ~1ms/reading
	if (shot.reject_shot) {
		shot.always_invalid = true;
	}
	if (shot.always_invalid) {
		return; // 120ms/190ms --> 60ms/112ms
	}

	// shot.soln boolean - whether or not the we have a calculated (x,y,v)
	// shot.valid - whether or not this shot gets shown to an ordinary user
	//
	shot.soln = shot.valid = (shot.source == "missed_shot") ||
	    !!shot.v && typeof(shot.x) == "number" && typeof(shot.y) == "number" ||
	    shot.res.method == "IMPACT";
	shot.invalid_reasons=[];
	
	if (!shot.valid) {
		shot.always_invalid = true;
	}

	if (shot.source == "missed_shot") {
		return; // no further processing
	}

	delete shot.screened_out_theta;
	delete shot.screened_out_phi;
	
	// various checks....

	// 15-May-2020: angles outside of "learned values" no longer marked invalid.

	// DMC: "angle checks" means to screen out shots on a global basis if (theta,phi) are out of bounds
	// recv_shot_invalidate_angles now DEPRECATED..... we now ALWAYS do angle checks:
	if (shot.valid) {
		// && typeof(shot.res.phi) != 'undefined') {
		var theta_nom = data.config.recv_shot_angle_theta_mean * 0.001 || 0;
		if (undefined == data.config.recv_shot_angle_theta_window ||
			// SOLO?
			(!data.config.recv_shot_angles_SOLO_fixed &&
				100 == data.config.recv_shot_angle_theta_window &&
				undefined == data.config.recv_shot_angle_theta_mean)
		) {
			data.config.recv_shot_angle_theta_window = 463;// 463 mils is 25 degrees
		}
		var theta_tol = data.config.recv_shot_angle_theta_window * 0.001 || 0; // 0 means don't screen
		// 2-Nov-2020: This change allowed ghost shots +/-463 mils  to happen if we have one mic pair with a
		//  late-firing, so we need to introduce a heuristic to try to filter them out.  Unfortunately this will
		// only work if "autolearn angles" has finished.  Screen only on side-angle for now.
		var theta_nom_learned=theta_nom;
		var theta_tol_learned=theta_tol;
		if (false === str.autolearn_angles && str.learned_angles) {
			theta_nom_learned=str.learned_angles.theta_mean;
			theta_tol_learned=str.learned_angles.theta_window*2;// be extra-forgiving, before we invalidate this shot
		}

		if (angle_outside(shot.res.theta,theta_nom,theta_tol)
		|| angle_outside(shot.res.theta,theta_nom_learned,theta_tol_learned)) {
			shot.screened_out_theta = true;
			shot.valid = false; shot.invalid_reasons.push("screened_out_sideangle");
			dbglog("DMC-1 recv_shot angle checks screened out 8-sensor ghost shot, theta=", shot.res.theta)
		}
		var phi_nom = data.config.recv_shot_angle_phi_mean * 0.001 || 0;;
		if (undefined == data.config.recv_shot_angle_phi_window ||
			(!data.config.recv_shot_angles_SOLO_fixed &&
				100 == data.config.recv_shot_angle_phi_window &&
				undefined == data.config.recv_shot_angle_phi_mean)
		) {
			data.config.recv_shot_angle_phi_window = 463;// 463 mils is 25 degrees
		}
		data.config.recv_shot_angles_SOLO_fixed = true;
		var phi_tol = data.config.recv_shot_angle_phi_window * 0.001 || 0;
		if (angle_outside(shot.res.phi,phi_nom,phi_tol)) {
			shot.screened_out_phi = true;
			shot.valid = false; shot.invalid_reasons.push("screened_out_plungeange");
			dbglog("DMC-3 recv_shot angle checks screened out 8-sensor ghost shot, phi=", shot.res.phi)
		}
	}

	var logical_mics_in_use = mic_mapping.logical_mics_in_use(mic_mapping.get_mic_mapping(shot)); // used below...
	// count how many mics fired
	var nmics_fired = 0;
	var missingmics=[];
	var zeromics=[];
	if (shot.mics && shot.mics.length) {
		for (var i in shot.mics) {
			if (shot.mics[i] >= 0) nmics_fired++;
			if (shot.mics[i] < 0 && logical_mics_in_use[i]) missingmics.push(parseInt(i)+1);
			if (shot.mics[i] == 0 && logical_mics_in_use[i]) zeromics.push(parseInt(i)+1);
		}
	}
	shot.whack_missing_mics = missingmics;
	shot.nmics_fired = nmics_fired;
	shot.is_starred = false; // starred for an ordinary user
	shot.is_starred_ro = false; // starred for RO/admin
	shot.star_reasons = [];
	
	// how many mic channels are available to us?
	var n_allmics = 0;
	if (shot.mic_mapping) {
		var avail_mics = mic_mapping.logical_to_phys(shot.mic_mapping,[1,2,3,4,5,6,7,8]);
		for (var i in avail_mics) {
			if (avail_mics[i] >= 0) n_allmics++;
		}
	//	dbglog("DMC3 nmics_fired=",nmics_fired,"n_allmics=",n_allmics,"mic_mapping=",shot.mic_mapping,"avail_mics=",avail_mics);
	}

	//  ---- "invalid" tests:

	var prev_shot = undefined, prev_shot_valid = undefined;
	if (shot_index == -1) {
		if (str.shots.length >= 1) {
			prev_shot = str.shots[str.shots.length-1]; // undefined if length < 1
			if (prev_shot.valid) prev_shot_valid = prev_shot;
		}
	} else { // Nov-2020 bugfix... search backwards for most recent previous shot reading that was valid:
		if (shot_index > 0) prev_shot = str.shots[shot_index-1];
		var i;
		for (i = shot_index-1; i >=0 && (prev_shot_valid = str.shots[i]) && prev_shot_valid.valid != true; i--) ;
		if (prev_shot_valid && prev_shot_valid.valid == false) prev_shot_valid = undefined;
	}

	//
	// optional: filter out duplicate shot.  LATER 20200116: ALWAYS do this
	if (/*data.config.recv_shot_invalidate_dup &&*/ shot.soln && prev_shot_valid && !shot.replayed_shot) {
//console.log("DMC VERIFY 1 shot_index=",shot_index,"shot.num=",shot.num,"prev_shot_valid.num=",prev_shot_valid.num)
		// console.log("DMC recv_shot_invalidate_dup str.shots.length=",str.shots.length,typeof shot_or_index,prev_shot_valid);
		var s2 = prev_shot_valid;
		if (s2.num == shot.num && s2.x == shot.x && s2.y == shot.y && s2.v == shot.v) {
			dbglog("DMC: invalidate duplicate measurement");
			// console.log("DMC calc_shot_valid invalidate_dup str.shots.length=", str.shots.length, typeof shot_or_index, prev_shot_valid);
			shot.score = 0; shot.is_starred = true;
			shot.valid = false; shot.invalid_reasons.push("dup_meas");
		}
	}
	
	if (shot.res.xy_valid != true) {
		shot.valid = false; shot.invalid_reasons.push("XY_VALID_FALSE"); // redundant, but let's label it
		shot.always_invalid = true;
	}

	// // optional: try to filter out BathurstMystery "tdup" shots
	// if (data.config.recv_shot_invalidate_tdup && shot.soln && prev_shot) {
	// 	var s2 = prev_shot;
	// 	if (Math.abs(shot.ts-s2.ts) < 300) {
	// 		dbglog("DMC: invalidate BathurstMystery t-duplicate measurement");
	// 		shot.valid = false; shot.invalid_reasons.push("BathurstMystery");
	// 	}
	// }

	// optional: try to filter out BathurstMystery shots in which mic2 is -1 and mic4 is 0 (or vice versa)

	// console.log("DMC n_allmics",n_allmics,"shot.mic_mapping",shot.mic_mapping,"soln",shot.soln);
	// console.log("DMC zeromics",zeromics,"len",zeromics.length);
	data.config.recv_shot_invalidate_bthst_m2m4 = true; // 20180810: always do this check
	// 2020-07-10 improve check: only invalidate if previous reading had all sensors fire
	if (data.config.recv_shot_invalidate_bthst_m2m4 && shot.soln && n_allmics == 5) {
		// console.log("test recv_shot_invalidate_bthst_m2m4",shot.mics);
		if ((shot.mics[2 - 1] < 0 && shot.mics[4 - 1] == 0)
			|| (shot.mics[4 - 1] < 0 && shot.mics[2 - 1] == 0)
		) {
		if (prev_shot && prev_shot.nmics_fired && prev_shot.nmics_fired == 5) {
				shot.valid = false; shot.invalid_reasons.push("BTHST_M2M4");
			}
		}
	}

	// invalidate shots in which >=2 mics are '0' and there is at least one missing mic (-1)
	if (data.config.recv_shot_invalidate_bthst_00_1 && shot.soln && n_allmics == 5) {
		// console.log("test recv_shot_invalidate_bthst_00_1", shot.mics,zeromics,zeromics.length);
		if (zeromics.length >= 2 && missingmics.length >= 1) {
			shot.valid = false; shot.invalid_reasons.push("BTHST_00_1");
		}
	}

	// optional: filter out ghost shot by speed:
	if (shot.valid && !(face && face.options && face.options.disable_ghost_check)) {
		// was: (data.config.recv_shot_invalidate_ghostV !== false) -- which will filter by speed if recv_shot_invalidate_ghostV === undefined
		var filter_by_speed = (data.config.recv_shot_invalidate_ghostV == true) && shot.valid && !(face && face.options && face.options.disable_ghost_check);
		if (n_allmics == 5) filter_by_speed = true; // always do this for 5-mic systems
		if (filter_by_speed) {
			var ghost_min = (face && face.options && face.options.ghost_min) || 300;
			var ghost_max = (face && face.options && face.options.ghost_max) || 1200;
			if (shot.v < ghost_min || shot.v > ghost_max) {
				dbglog("DMC: invalidate ghost shot, based on speed");
				shot.valid = false; shot.invalid_reasons.push("ghost_V");
			}
		}
	}

	data.config.recv_shot_invalidate_need_three_chronos = true; // 20180810: always do this check
	if (data.config.recv_shot_invalidate_need_three_chronos && shot.soln && n_allmics >= 6) {
		var n_chronos = 0;
		if (shot.mics[1 - 1] >= 0 && shot.mics[3 - 1] >=0) n_chronos++;
		if (shot.mics[2 - 1] >= 0 && shot.mics[6 - 1] >=0) n_chronos++;
		if (shot.mics[4 - 1] >= 0 && shot.mics[7 - 1] >=0) n_chronos++;
		if (shot.mics[5 - 1] >= 0 && shot.mics[8 - 1] >=0) n_chronos++;
		if (n_chronos < 3) {
			shot.valid = false; shot.invalid_reasons.push("NEED_3_CHRONOS");
		}
	}

//	// for 8-mic systems using standard geometry (and therefore method = "ITER"),
//	// invalidate shots missing a single mic; this is a likely symptom of a weak signal,
//	// which may cause an invalid reading on one or more of the fired mics
//	//
//	// later: recv_shot_invalidate_7mics is no longer used and should always be false
//	if (data.config.recv_shot_invalidate_7mics && shot.res.method == "ITER" && n_allmics == 8 && nmics_fired !=6 && nmics_fired != 8) {
//		shot.valid = false; shot.invalid_reasons.push("ONLY_6_OR_8");
//	}

	// for 8-mic systems using standard geometry (and therefore method = "ITER"),
	// invalidate shots if any chrono shows a subsonic readings - this is impossible,
	// and suggests a weak signal causing an invalid reading
	if (shot.res.method == "ITER" && n_allmics == 8) {
		function is_subsonic(v,a) {
			// console.log("is_subsonic",v,a,v !== undefined && Math.abs(v) < a);
			return v !== undefined && Math.abs(v) < a;
		}
		if (is_subsonic(shot.res.raw_v13,shot.res.soundspeed)) {
			shot.is_starred_ro = true; shot.star_reasons.push("SUBSONIC_CHRONO v13=" + trim_num(shot.res.raw_v13,0));
		}
		if (is_subsonic(shot.res.raw_v26,shot.res.soundspeed)) {
			shot.is_starred_ro = true; shot.star_reasons.push("SUBSONIC_CHRONO v26=" + trim_num(shot.res.raw_v26,0));
		}
		if (is_subsonic(shot.res.raw_v47,shot.res.soundspeed)) {
			shot.is_starred_ro = true; shot.star_reasons.push("SUBSONIC_CHRONO v47=" + trim_num(shot.res.raw_v47,0));
		}
		if (is_subsonic(shot.res.raw_v58,shot.res.soundspeed)) {
			shot.is_starred_ro = true; shot.star_reasons.push("SUBSONIC_CHRONO v58=" + trim_num(shot.res.raw_v58,0));
		}
	}


	// 15-May-2020: we now "star" shots with angles outside learned values
	if (false === str.autolearn_angles && str.learned_angles && shot.valid) {
		// if we have learned acceptable angles, set shot.valid accordingly
		// screen incoming shots accoring to learned angles
		var theta_nom=str.learned_angles.theta_mean;
		var theta_tol=str.learned_angles.theta_window; // 0 means don't screen
		// for nonzero theta_window, screen:


		if (angle_outside(shot.res.theta,theta_nom,theta_tol)) {
			shot.is_starred = true; shot.star_reasons.push("sideangle diff " + trim_num(1000*(shot.res.theta-theta_nom),0));
			dbglog("DMC-1 recv_shot autolearn starred theta=", shot.res.theta,theta_nom,theta_tol)
		}
		var phi_nom= str.learned_angles.phi_mean;
		var phi_tol = str.learned_angles.phi_window;
		if (angle_outside(shot.res.phi,phi_nom,phi_tol)) {
			shot.is_starred = true; shot.star_reasons.push("plungeangle diff " + trim_num(1000*(shot.res.phi-phi_nom),0));
			dbglog("DMC-3 recv_shot autolearn starred phi=", shot.res.phi,phi_nom,phi_tol)
		}
	}

//	// optional: filter out shots missing at least two mics
//	if (data.config.recv_shot_invalidate_missing2) {
//		if (missingmics.length >= 2) { shot.valid = false; shot.invalid_reasons.push("missing2"); }
//	}
//	// console.log("DMC calc-shot-valid-2",data.config.recv_shot_invalidate_missing2,missingmics.length,shot.valid)
//
//	// optional: filter out shots missing at least three mics
//	if (data.config.recv_shot_invalidate_missing3) {
//		if (missingmics.length >= 3) { shot.valid = false; shot.invalid_reasons.push("missing3"); }
//	}
//
//	// optional: filter out shots missing at least four mics
//	if (data.config.recv_shot_invalidate_missing4) {
//		if (missingmics.length >= 4) { shot.valid = false; shot.invalid_reasons.push("missing4"); }
//	}

	if (data.config.star_shot_vneg && (
		shot.res.raw_v13 < 0 ||
		shot.res.raw_v26 < 0 ||
		shot.res.raw_v47 < 0 ||
		shot.res.raw_v58 < 0
	) ) {
		shot.valid = false;
		shot.invalid_reasons.push("negative speed in v13, v26, v47 or v58");
	}

	// SOLO: recv_shot_show=[on,close,all]
	if (shot.valid) {
		shot.debug_recv_shot_show = data.config.recv_shot_show;
		if (data.config.recv_shot_show == 'all'){ // show everything
			; // accept all shots
		} else if (data.config.recv_shot_show == 'close') {
			if (face && face.centrecoords && face.targetboard
				&& shot.x > face.centrecoords.x - face.targetboard.width/2 - 1000
				&& shot.x < face.centrecoords.x + face.targetboard.width/2 + 1000
				&& shot.y > face.centrecoords.y - face.targetboard.height/2 - 2000
				&& shot.y < face.centrecoords.y + face.targetboard.height/2 + 3000
			) {
				; // shot is on target; accept;
			} else {
				shot.valid = false; shot.invalid_reasons.push("recv_shot_show not 'close'");
				; // shot is not close enough; invalidate
				console.log(shot.x, face.centrecoords.x - face.targetboard.width/2 - 1000
				,shot.x > face.centrecoords.x - face.targetboard.width/2 - 1000);
				console.log(shot.x, face.centrecoords.x + face.targetboard.width/2 + 1000
				);
				console.log(shot.y, face.centrecoords.y - face.targetboard.height/2 - 2000
				);
				console.log(shot.y, face.centrecoords.y + face.targetboard.height/2 + 3000
				);

			}
		} else { // default to 'on', only show on-target shots
			if (face && face.centrecoords && face.targetboard
				&& shot.x > face.centrecoords.x - face.targetboard.width/2
				&& shot.x < face.centrecoords.x + face.targetboard.width/2
				&& shot.y > face.centrecoords.y - face.targetboard.height/2
				&& shot.y < face.centrecoords.y + face.targetboard.height/2) {
				; // shot is on target; accept;
			} else {
				shot.valid = false; shot.invalid_reasons.push("recv_shot_show not 'on'");
				; // shot is not on target; invalidate
			}
		}
	}

	//  ---- on/off controllable "*" tests

	// mark the shot with a '*' if anything is questionable
	if (data.config.star_shot_missing_mic && shot.whack_missing_mics && shot.whack_missing_mics.length > 0) {
		shot.is_starred_ro = true;
		shot.star_reasons.push('missing mics '+shot.whack_missing_mics);
	}
	
	if (data.config.star_shot_xyerr) {
		var xyerr_limit = data.config.star_shot_xyerr_amt || 25;
		if (shot.res && shot.res.xy_error > xyerr_limit) {
			shot.is_starred = true;
			shot.star_reasons.push("XY-err "+(trim_num(shot.res.xy_error,0)));
		}
	}

	if (data.config.star_shot_verr) {
		var v_limit = data.config.star_shot_verr_amt || 1;
		if (shot.res && shot.res.v_error > v_limit) {
			shot.is_starred = true;
			shot.star_reasons.push("V-err "+(trim_num(shot.res.v_error,0)));
		}
	}

}

function calc_shot_reject(shot,str) { // set shot.reject if criteria are met
	// debugging:
	var dbglog = console.log;
	// no debug:
	if (!data.config.option_debuginfo) dbglog = function() {};

	delete shot.reject_shot;

	// string full, too many readings?
	var max_shots_length = 500; // hardcode for now
	if (str.shots.length >= max_shots_length ) {
		dbglog("calc_shot_reject: shots.length exceeded",max_shots_length);
		shot.reject_shot = true;
		// crud this out
		var crud = { strings: {} };
		crud.strings[str._id] = {};
		crud.strings[str._id].shots_full = true;
		apply_crud(crud);
	}

	var nmics_fired = shot.nmics_fired;

	// reject single mic taps?
	if (nmics_fired < 2 && data.config.recv_shot_reject_tap) {
		dbglog("calc_shot_reject: single-tap");
		shot.reject_shot = true;
	}

	// reject multi-mic firings?
	if (nmics_fired > 1 && !shot.soln && data.config.recv_shot_reject_multi) {
		dbglog("calc_shot_reject: multi-mic readings w/o a solution");
		shot.reject_shot = true;
	}

	// reject near/far BROKEN for now - need result from calc_scorezone(); LATER
	//
	// reject shots that have missed the target, but are near (within 2 targetboard from centre)?
	if (data.config.recv_shot_reject_off_near && shot.soln && shot.value == "off"
		&& face && face.centrecoords && face.targetboard
		&& shot.x > face.centrecoords.x - face.targetboard.width * 2
		&& shot.x < face.centrecoords.x + face.targetboard.width * 2
		&& shot.y > face.centrecoords.y - face.targetboard.height * 2
		&& shot.y < face.centrecoords.y + face.targetboard.height * 2) {
			dbglog("calc_shot_reject: off-target near shot");
			shot.reject_shot = true;
	}

	// reject shots that have missed the target and are far away?
	if (data.config.recv_shot_reject_off_far && shot.value == "off"
		&& face && face.centrecoords && face.targetboard
		&& (
			shot.x < face.centrecoords.x - face.targetboard.width * 2
			|| shot.x > face.centrecoords.x + face.targetboard.width * 2
			|| shot.y < face.centrecoords.y - face.targetboard.height * 2
			|| shot.y > face.centrecoords.y + face.targetboard.height * 2
			)
		) {
		dbglog("calc_shot_reject: off-target far");
		shot.reject_shot = true;
	}
}

function timer_expiredQ(timer) { 
	return timer && timer.running && Date.now() > timer.endtime;
}

function global_timers_shot_invalid() {
	var global_timers = global.data.config.global_timers;
	if (!global_timers || !global_timers.length) return false;
	// check shot against global timers.
	// if there are *any* global timers, then a shot will be struck out UNLESS
	// it occurs while a timer is running.
	for (var i in global_timers) {
		var timer = global_timers[i];
		if (timer.running) { // could be: active or finished.
			if (!timer_expiredQ(timer)) {
				return false;
			}
		}
	}
	
	return true; // shot is out-of-bounds; strike it?
}

var gpioFeature = require('./gpio.js');

function recv_shot_assign_shooter(shot,str) {

	var shooter_count = str.shooter_count || 1;
	var face = data.faces[str.face_id];

	if (!shot.valid) {
		shot.value = "-"; shot.score = 0;
		return;
	}

	if (global_timers_shot_invalid()) { // TODO: make this optional/controllable?
		console.log("DMC kluge, strikeout");
		shot.strikeout = true;
	}



	// assign valid shots to shooters

	// default
	shot.shooter_num = shooter_count - 1;

	// increment for multiple shooters
	if (shooter_count > 1) {
		for (var i = str.shots.length - 1; i >= 0; i--) {
			var shot_last = str.shots[i];
			if (shot_last.valid && !shot_last.strikeout) {
				if (shot_last.shooter_num > 0) {
					shot.shooter_num = shot_last.shooter_num - 1;
				}
				break;
			}
		}
	}

	// see if shot should be assigned to another shooter if current shooter's string is full
	if (data.config.shots_mode > 0 && shooter_count > 1) {

		// count shots for each shooter
		var count = [0, 0, 0];
		for (var i in str.shots) {
			var s2 = str.shots[i];
			if (s2.valid && !s2.sighter && !s2.strikeout) count[s2.shooter_num]++;
		}

		// check each shooter starting with assumed assignment
		for (var i = 0; i < shooter_count; i++) {
			var cur = shot.shooter_num - i;
			if (cur < 0) cur += shooter_count;

			// this shooter has room for the shot
			if (count[cur] < data.config.shots_mode) {
				shot.shooter_num = cur;
				break;
			}

		}
	}

	// whether should be fclass scoring
	var use_fclass = face.scoringzones_f && str.fclass_target && str.fclass_target[shot.shooter_num];
	// use bullet diameter if not standard
	if (face.bullet_dia) shot.bullet_dia = face.bullet_dia;

	// whether user has custom bullet diameter
	var user_bullet = false;
	if (data.config.use_role_shooter && str.role_shooter && str.role_shooter[shot.shooter_num]) {
		var user = data.users[str.role_shooter[shot.shooter_num]];
		if (user && user.bullet_dia) {
			shot.bullet_dia = user.bullet_dia;
			user_bullet = true;
		}
	} else if (str.bullet_dia) {
		shot.bullet_dia = str.bullet_dia;
		user_bullet = true;
	}

	//dbglog("DMC4- shot.valid, score value = ", shot.value, "x,y=",shot.x,shot.y,"use_fclass=",use_fclass,"user_bullet=",user_bullet);
	var z = calc_scorezone(shot, face, use_fclass);
	shot.value = z.name;
	shot.score = z.score;
	// hooks needed to implement "fly" rules:
	// 1 - auto-conversion of sighters
	// 2 - assignment of group-bonus
	// fly-hook
	if ('function' == typeof face.hookfn_recv_shot_assign_shooter) {
		// note that we are extrememly early in the shot assignment process.  For example,
		// we have not yet calculated group size.
		// set shot.sighter in accordance with "fly" rules:
		face.hookfn_recv_shot_assign_shooter(str, shot, face, use_fclass); // will not alter shot.value, shot.score
		// later, calc_string_score will calculate bonus points for group size
	}

	gpioFeature.triggerByZone(z);

	// check and set if shot should be auto sighter
	set_shot_auto_sighter(str, shot);

//dbglog("DMC4 shot.valid, score value = ", shot.value, "x,y=",shot.x,shot.y,"use_fclass=",use_fclass,"user_bullet=",user_bullet);

}





// temporary - cloned from client_face.js...
// TODO: scoring of all types of zones
// TODO-soon: F-Class vs. TR scoring
function calc_scorezone(shot, face, fclass) {
//log4.debug("DMC calc_scorezone GPIO18 check",shot.GPIO18,"1"== shot.GPIO18,1==shot.GPIO18,'valid:',shot.valid);
	var start = _performance();

	// invalid shot
	if (!shot.valid) {
		// console.log('calc_scorezone !shot.valid' + _performance(start) + 'ms)');
		return { name: "-", score: 0 };
	}
	
	// special method, for now at least, for "IMPACT":
	if (shot.res.method == "IMPACT") {
		var result = { name: "-", score: 0 }
		for (var i in shot.res.impact_mics) {
			if (shot.res.impact_mics[i] != -1 && result.name == "-") {
				// report first fired mic
				var mic_number = +i +1;
				result.name = "" + mic_number;
				result.score = mic_number;
			}
		}
		return result;
	}

	var x = shot.x, y = shot.y;
	
	// shot can define the bullet_dia
	var use_bullet = shot.bullet_dia || 7.82;
	
	// choose zones
	var zones = !fclass ? face.scoringzones : face.scoringzones_f;
	if (!zones) {
		// console.log('calc_scorzone !zones' + _performance(start) + 'ms)');
		return { name: "0", score: 0 };
	}

	
	/* TODO: 
	// search "subzones" first
	if (typeof ctarget.subzones == 'object' && ctarget.subzones.length > 0) {
		var n = ctarget.subzones.length;
		for (i = 0; i < n; i++) {
			var sub_ctarget = target_stuff.targetlist[ctarget.subzones[i].tfacename];
			var z = calc_shot_scoringzone(sub_ctarget,x,y);
			if (z) return z;
		}
	}
	*/
	
	// search scoring areas, in forward order
	for (var i = 0; i < zones.length; i++) {
		var zone = zones[i];
		if (is_shot_in_scoringzone(x, y, zone, use_bullet)) {
			// console.log('calc_scorzone foundzone' + _performance(start) + 'ms)');
			return zone;
		}
	}
	
	// default: valid shot near targetboard, no zone
	// console.log('calc_scorzone near-no-zone' + _performance(start) + 'ms)');
	return { name: "0", score: 0 };
}


// true/false: is (x,y) inside scoring zone 'z'?
function is_shot_in_scoringzone(x,y,z /* zone */ ,bullet_dia) {
    // handle z == null or malformed
    if (!z || !z.type) return false;

    var centre = z.centre ? z.centre : {x:0, y:0};
    // each zone may override the global definition of bullet diameter
    var my_bullet_dia = typeof z.bullet_dia != 'undefined' ? z.bullet_dia : bullet_dia;
    my_bullet_dia = my_bullet_dia || 0; // undef->0

    if (z.type.name == "ring") {
      var radius = z.type.radius ? z.type.radius : z.type.dia/2;
      var shot_r = Math.sqrt(
        Math.pow(x-centre.x,2)+
        Math.pow(y-centre.y,2));
      if (shot_r <= radius + (my_bullet_dia/2)) { // DMC - include bullet dia. --> OK if negative!
        return true;
      }
    } else if (z.type.name == "rect") { // TBD: should bullet dia be incorporated here?
      if (x >= centre.x-z.type.width/2 &&
          x <= centre.x+z.type.width/2 &&
          y >= centre.y-z.type.height/2 &&
          y <= centre.y+z.type.height/2) {
        return true;
      }
    } else if (z.type.name == "poly") {
      // we may have entered "shifted" vertices for convenience; shift these to where
      // they belong
      var sx = z.type.shiftamt ? z.type.shiftamt[0] : 0;
      var sy = z.type.shiftamt ? z.type.shiftamt[1] : 0;

      if (point_in_poly([x-sx,y-sy], z.type.vertices)) {
        return true;
      }
    } else if (z.type.name == "and") { // an array of subzones; the shot must be in ALL of them;
      var jj;

      for (jj = 0; jj < z.type.members.length; jj++) {
        var zzz = z.type.members[jj]; // could be null
        if (!is_shot_in_scoringzone(x,y,zzz,bullet_dia)) { // first false is all we need
          return false;
        } // if
      } // for
      // all subconditions passed
      return true;
      // note: degenerate case: will return TRUE if no nonempty elements
    } else if (z.type.name == "or") { // an array of subzones; the shot must be in ANY of them;
      var jj;
      for (jj = 0; jj < z.type.members.length; jj++) {
        var zzz = z.type.members[jj]; // could be null
        if (zzz && is_shot_in_scoringzone(x,y,zzz,bullet_dia)) { // first true is all we need
          return true;
        }
      } // for
      // all subconditions failed
      return false;
      // note: degenerate case: will return FALSE if no nonempty elements
    } else if (z.type.name == "not") { // one subzones; the shot must NOT be in it
      return !is_shot_in_scoringzone(x,y,z.type.member,bullet_dia);
    } else { // unrecognized type
      // fall-through, and return false
    }
  return false;
}


function point_in_poly(point, vs) {
    // ray-casting algorithm based on
    // http://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/pnpoly.html
    
    var x = point[0], y = point[1];
    
    var inside = false;
    for (var i = 0, j = vs.length - 1; i < vs.length; j = i++) {
        var xi = vs[i][0], yi = vs[i][1];
        var xj = vs[j][0], yj = vs[j][1];
        
        var intersect = ((yi > y) != (yj > y))
            && (x < (xj - xi) * (y - yi) / (yj - yi) + xi);
        if (intersect) inside = !inside;
    }
    
    return inside;
};


module.exports.calc_scorezone = calc_scorezone;


module.exports.update_string_stats_autolearn =

function update_string_stats_autolearn(string_id) {
	log("update_string_stats_autolearn", string_id);
	var start = _performance();
	// debugging:
	var dbglog = console.log;
	// no debug:
	if (!data.config.option_debuginfo) dbglog = function() {};
dbglog('update_string_stats_autolearn start - ', _performance(start)+'ms');

	var str = data.strings[string_id];
	if (!str) return;
	
	var face = data.faces[str.face_id];
	if (!face) return;
	
	var crud = {};
	crud.strings = {};

	var my_crud_string = crud.strings[str._id] = {
		shots:[],
	};
	
// start autolearn

	var n_valid_unstruck = 0;
	var i;
	for (i=0; i < str.shots.length; i++) { // TODO: this is O(n^2), can we do anything about this?
		var s = str.shots[i];
        if ('object' != typeof s) continue; // robustify
		var before_is_starred = s.is_starred, before_is_starred_ro = s.is_starred_ro, before_is_valid = s.valid;
		calc_shot_valid(i,str); // DMC 30 May; shot.valid, shot.is_starred might change
		//dbglog("DMC update_string_stats",i,before_is_starred,"==>",s.is_starred,before_is_valid,'==>',s.valid);
		var crud_shot = my_crud_string.shots[i];
		if (!crud_shot) crud_shot = my_crud_string.shots[i] = {};
		if (before_is_starred != s.is_starred) { crud_shot.is_starred = s.is_starred; crud_shot.star_reasons = s.star_reasons;}
		if (before_is_starred_ro != s.is_starred_ro) { crud_shot.is_starred_ro = s.is_starred_ro; crud_shot.star_reasons = s.star_reasons;}
		if (s.strikeout_recalc_autolearn) {
			crud_shot.strikeout_recalc_autolearn = false;
			my_crud_string.autolearn_angles = str.autolearn_angles = true;
//crud_shot.strikeout_recalc_autolearn_done = true;
dbglog("DMC update_string_stats_autolearn strikeout_recalc_autolearn");
		}
		if (before_is_valid != s.is_valid) {
			crud_shot.valid = s.valid; crud_shot.invalid_reasons = s.invalid_reasons;
			// console.log("DMC update_string_stats",crud_shot);
		}
		if (s.valid && !s.strikeout) n_valid_unstruck++;
	}

dbglog("DMC update_string_stats str.autolearn_angles",str.autolearn_angles,"n_valid_unstruck",n_valid_unstruck, _performance(start)+'ms');

	// autolearn
	if (str.autolearn_angles && n_valid_unstruck > 2) {
		var angle_stats = calc_string_angles(str);
		/* { v_avg: v_avg, v_sd: v_sd
		,th_avg:th_avg, th_sd:th_sd
		,ph_avg:ph_avg, ph_sd:ph_sd
		,n_valid:count }
		*/
		var stable_angle = 20 * 0.001; // "stable" if angles SD < 20
		if (angle_stats.th_sd < stable_angle && angle_stats.ph_sd < stable_angle) {
			my_crud_string.autolearn_angles = false; // we are done learning
			my_crud_string.learned_angles = {
				theta_mean: (+trim_num(angle_stats.th_avg,3)), // argh!! .toFixed() returns STRING!!
				phi_mean: (+trim_num(angle_stats.ph_avg,3)),
				theta_window: stable_angle*3, // 3x SD
				phi_window: stable_angle*3*2, // plunge angle can be much more variable, for flopping target frames
			};
		}
		dbglog("DMC autolearn_angles now",my_crud_string.autolearn_angles," stats=", angle_stats, my_crud_string.autolearn_angles, my_crud_string.learned_angles, _performance(start)+'ms');
	}
	
	// end autolearn
	
	apply_crud(crud);
	dbglog("DMC autolearn_angles done ", _performance(start)+'ms');
}



// handle whether this new shot should be considered a sighter
// data.config.sighter_enabled: DEPRECATED
// sighter_count: ==0 for no sighters, > 0 for limited sighters, or -1 for unlimited
// sighter_convertible: true/false
// str.on_record[]: if true for the current shooter, don't assign more shots
function set_shot_auto_sighter(str, shot) {

	if (!shot.valid) return;
	
	if (data.config.sighter_count == 0) return;
	
	// current shooter is on record, don't make sighters
	if (str.on_record && str.on_record[shot.shooter_num]) return;
	// count sighters already on the currently shooter against the limit
	
	var count = 0;

	// check all shots for current shooter in string
	for (var i in str.shots) {
		var s2 = str.shots[i];
		if (!s2.valid || s2.shooter_num != shot.shooter_num) continue;
		if (s2.strikeout) continue;
		
		if (s2.sighter == true) count++;
		else return; // found a non-sighter, exit
		
		// limited sighters and too many already
		if (data.config.sighter_count > 0 && count >= data.config.sighter_count) return;

	}
	
	// passed the test: should be a sighter
	shot.sighter = true;

}




function calc_mean_sd(samples) { // cloned from smt.common.lib/public/javascripts/utils.js
	if (samples.length < 1) return {mean: 0, sd: 0};
	var _avg = 0;
	var _sd = 0;
	for (i in samples) {
		_avg += samples[i];
	}
	_avg /= samples.length;
	for (i in samples) {
		_sd += (samples[i] - _avg)*(samples[i] - _avg);
	}
	_sd = Math.sqrt(_sd / samples.length);
	return {mean:_avg, sd: _sd};
}

function calc_string_angles(string_data) { // all valid not-struck-out shots, all shooters
	var count = 0;
	var shooter_count = string_data.shooter_count || 1;
	
	var speeds = [], th=[], ph=[];

	var shots = string_data.shots;
	for (var i in shots) {
		var shot = shots[i];
		if (!shot || !shot.valid /* || shot.sighter */ || shot.strikeout) continue;
		count++;
		
		speeds.push(shot.v);
		if (shot.res && undefined != shot.res.theta) th.push(shot.res.theta);
		if (shot.res && undefined != shot.res.phi) ph.push(shot.res.phi);

	}
	var ss = calc_mean_sd(speeds); // calc theta stats
	var v_avg = ss.mean, v_sd = ss.sd;

	var ss = calc_mean_sd(th); // calc theta stats
	var th_avg = ss.mean, th_sd = ss.sd;
	
	var ss = calc_mean_sd(ph); // calc phi stats
	var ph_avg = ss.mean, ph_sd = ss.sd;

	return { v_avg: v_avg, v_sd: v_sd
		,th_avg:th_avg, th_sd:th_sd
		,ph_avg:ph_avg, ph_sd:ph_sd
		,n_valid:count
	};
}

function trim_num(nn,ndigs) { // type-safe .toFixed()
		if (typeof ndigs == 'undefined' || ndigs < 0) ndigs = 0;
		if (typeof nn == 'number') return +nn.toFixed(ndigs); // '+' needed to return 'number' not 'string'
		return nn;
}


