/*******************************************************************************
 viewer_shotlist.js: client module for sidebar shots list

 $Author: dmc $
 $Revision: 3206 $
 $Date: 2023-07-06 10:53:57 -0300 (Thu, 06 Jul 2023) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer_shotlist.js $
*******************************************************************************/

// set early update(), for display only
var cur_shooter = 0; // which shooter to view (0 = left)

shotlist = new function() {

	this.init = function() {
		this.div = $("#divShotList");

		// reposition shot info box when scrolling list
		this.div.bind("scroll", shotbubble.update_position);
		
		this.string_info_timer = setInterval(this.update_string_info, 1000);

	}
	
	this.clear = function() {
		hide(this.div);
		this.div.empty();
		set_pos(this.div, "height", 100); // allow to shrink needed to rescale table
		this.focus_tr = null;
	}
	
	this.focus_tr = null; // table row of focus shot in list
	
	this.shot_details_timer;

	this.table_touch_start = 0; // for controlling scrolling of the shots table
	
	this.string_info_div = null;

	// auto updating string text block (short term solution)
	this.update_string_info = function() {
		var div = this.string_info_div;
		if (!div) return;
		
		div.empty();
		
		if (!cur_string()) return; // can happen since timer starts immediately

		var sdata = cur_string().data;
		
		// use event details (for live and review mode)
		var str = "";
		if (live_mode() && cur_event()) {
			str = "<b>" + g_config.event_data.event_str + "</b><br>"
						+ g_config.event_data.match_str + "<br>" + g_config.event_data.stage_str;
		} else if (!live_mode() && sdata.event_data
		 && sdata.event_data.event_id >= 0
		 && sdata.event_data.match_id >= 0
		 && sdata.event_data.stage_id >= 0) {
			str = "<b>" + sdata.event_data.event_str + "</b><br>"
						+ sdata.event_data.match_str + "<br>" + sdata.event_data.stage_str;
		}
		if (str) {
			var inner = $("<div>").append(str).css("font-size", "14px");
			div.append(inner);
			div.append("<hr>");
		}

		// frame and face name
		var str = "<b>" + cur_string().frame().data.name + "</b>";

		var device = cur_string().frame().device() &&
			cur_string().frame().device().data || undefined;
		var rfSignal = device && device.remote_rf_level || undefined; // could be null
		if ( rfSignal !== undefined /* && !toolbar_military.is_mil()*/) { // DMC edit 8 July 2022
			str += " [" + rfSignal + "]";
		}
		
		// ping health?
		var ping_health = device && device.ping_health;
		if ( ping_health && ping_health.show && g_config.is_health) {
			str += ("&nbsp;<span title='"+JSON.stringify(ping_health,null,2)+"' >" + ping_health.show+'</span>') ;
		}

		var facename = cur_string().face().data.name;
		if (typeof global_customFaceName === "function") {
			facename = global_customFaceName(facename);
		}
		str += "<br><span class='shotlist-face-name'>" + facename + '</span>';
		if (str) {
			var inner = $("<div>").append(str);
			div.append(inner);
		}

		// string stats
		var str = "";
		
		if (!printing && !live_mode()) {
			str += "<br>" + sdata.name;
		}
		
		if (g_config.option_group_stats) {
            var unit = get_coords_unit();
            var mpier = get_coords_mpier();
            var ndigs = get_coords_ndigs();

			var stats = sdata;
			if (sdata.shooterStats && sdata.shooterStats[cur_shooter]) {
				var stats = sdata.shooterStats[cur_shooter];
			}
			if (stats && stats.use_string_stats) {
				if (toolbar_military.is_mil()) {
					str += "<hr>ES - " + (stats.stats_group_size * mpier).toFixed(ndigs) + unit;
				} else {
					str += "<hr>Group: " + (stats.stats_group_size * mpier).toFixed(ndigs) + " " + unit;
					str += "<br>(" + (stats.stats_width * mpier).toFixed(ndigs) + "w x ";
					str += (stats.stats_height * mpier).toFixed(ndigs) + "h)";
				}
			}
		}
		// DMC 10 Sep 2017
		var score = calc_string_score(sdata, cur_shooter);
		if (g_config.option_xtrastats) {
			if ('number'==typeof score.th_avg) str += "<br>th="+(score.th_avg*1000).toFixed(0)+"/"+(score.th_sd*1000).toFixed(0);
			if ('number'==typeof score.ph_avg) str += "<br>ph="+(score.ph_avg*1000).toFixed(0)+"/"+(score.ph_sd*1000).toFixed(0);
			if (sdata.sim_errors) {
				var se = sdata.sim_errors;
				str += '<br>green '+se.n_green+' '+(se.err_tot_green).toFixed(0);
				str += '<br>yellow '+se.n_yellow+' '+(se.err_tot_yellow).toFixed(0);
				str += '<br>red '+se.n_red+' '+(se.err_tot_red).toFixed(0);
				str += '<br>orange '+se.n_orange;
				str += '<br>perm_mics '+se.perm_mics;
				console.log(sdata.sim_errors);
			}
		}
		if (str) {
			var inner = $("<div>").append(str).css("font-size", "14px");
			if (toolbar_military.is_mil()) inner.css("font-size", "24px");
			div.append(inner);
		}

	}

	// select shot in table to focus
	// used by both types of shot tables
	this.user_press_shot = function(i) {
		cur_string().select_shot(i); // will set shotbubble.enabled correctly
		var scroll = this.div[0].scrollTop; // return to the same scroll position
		update();
		this.div.scrollTop(scroll);
		set_shot_selection_timer();
	}

	this.bullet_dia_list = [
		{name:"223",dia:5.56},
		{name:"6mm",dia:6},
		{name:"6.5",dia:6.5},
		{name:"7mm",dia:7},
		{name:"308",dia:7.82},
		{name:"338",dia:8.58},
	];

	// build the sidebar (string info, shots table)
	this.update = function() {
		
		this.clear();
		if (!cur_string()) return;

		//log("shotlist.update()");

		// shortcuts
		var string = cur_string();
		var shooter_count = string.data.shooter_count || 1;

		if (cur_shooter >= shooter_count) cur_shooter = 0; // to handle switching from 3 to 2

		// auto select a shooter and check cur user
		var cur_user_shooter_index = user_is_shooter(); // cur_string, cur_user
		if (live_mode() && cur_user_shooter_index >= 0) cur_shooter = cur_user_shooter_index; // lock

		// string header panel container
		var panel = $("<div id='divStringPanel' class='dark_overlay'>");
		this.div.append(panel);
		
		// allow touch scrolling only when scrolling is possible
		// to prevent scrolling the entire page itself
		//this.div.bind("touchstart", function(e) { e.originalEvent.allow_touchstart = true; });
		//this.div.bind("touchmove", function(e) { e.originalEvent.allow_touchmove = true; });
		this.div.bind("touchstart", function(e) {
			shotlist.table_touch_start = e.originalEvent.touches[0].pageY;
			e.originalEvent.allow_touchstart = true;
		});
		this.div.bind("touchmove", function(e) {
			var delta = e.originalEvent.touches[0].pageY - shotlist.table_touch_start;
			if (delta > 0 && shotlist.div[0].scrollTop == 0) return;
			if (delta < 0 && shotlist.div[0].scrollTop == shotlist.div[0].scrollHeight - shotlist.div[0].clientHeight) return;
			e.originalEvent.allow_touchmove = true;
			set_shot_selection_timer();
		});

		var frame = string.frame();
		if (printing) {
			panel.css("background-color", "#FFF");
			panel.css("color", "black");
		} else if (!live_mode() || !frame.data.enabled) {
			panel.css("background-color", "#666");
			panel.css("color", "white");
		} else if (!frame.device()) {
			panel.css("background-color", "#FF9");
			panel.css("color", "black");
		} else if (!frame.data.connected || !g_connected) {
			panel.css("background-color", "#F99");
			panel.css("color", "black");
		} else {
			panel.css("background-color", "#AFA");
			panel.css("color", "black");
		}
		
		this.string_info_div = $("<div>");
		panel.append(this.string_info_div);

		this.update_string_info();
		
		if (!printing && live_mode() && user_can_save_string()) {

			// determine whether scores are final and shooters assigned
			var scores_are_final = g_config.shots_mode > 0;
			if (scores_are_final && g_config.use_role_scorer && !string.data.role_scorer) {
				scores_are_final = false;
			}
			for (var i = 0; i < shooter_count && scores_are_final; i++) {
				if (g_config.shots_mode > 0) {
					var score = calc_string_score(string.data, i);
					if (score.count != g_config.shots_mode) {
						scores_are_final = false;
						break;
					}
				}
				if (g_config.use_role_shooter && (!string.data.role_shooter || !string.data.role_shooter[i])) {
					scores_are_final = false;
					break;
				}
			}
			
			// mode where shooters must accept their score
			var string_is_accepted = scores_are_final;
			if (scores_are_final && g_config.option_must_accept && g_config.use_role_shooter) {
				
				var count_accepted = 0;
				if (string.data.accepted) {
					for (i in string.data.accepted) {
						if (string.data.accepted[i]) {
							count_accepted++;
						}
					}
				}
				
				if (count_accepted != shooter_count) {
					string_is_accepted = false;
				}
				
			}
			
			// finish string button
			var show_save_button = true;
			var save_text = S("STRING_FINISH");
			var save_bold = true;
			
			// we care about scores being final
			if (g_config.shots_mode > 0) {
				var save_text = "Save";
				var save_bold = false;
				show_save_button = admin_check() || ro_check() || g_config.option_access_mode == 0; // who can save non-final?
				if (string_is_accepted) {
					save_text = "Save<br>Final<br>Scores";
					save_bold = true;
					show_save_button = true;
				}
				if (string.data.note) {
					save_text = "Save<br>with<br>Note";
					save_bold = true;
				}
			}
			
			if (g_config.basic_mode) { // override "save" button behaviour
				save_text = "Save";
				save_bold = true;
				show_save_button = true;
			}

			if (toolbar_military.is_mil()) {
				save_text = "Save/Clear";
				save_bold = true;
				show_save_button = true;
			}

			if (g_config.is_NRA) {
				show_save_button = admin_check() || ro_check(); // who can save non-final?
//				show_save_button = false;
			}

			if (typeof global_customShotlistButtons === "function") {
				show_save_button = global_customShotlistButtons(panel, show_save_button);
			}

			if (show_save_button) {
				
				var item = $("<button id='cmdNewString' class='wide_button'>");
				item.html(save_text);
				item.css("width","80%");
				item.css("margin-top","5px");
				item.css("margin-bottom","5px");
				if (save_bold) item.css("font-weight","bold");
				
				var mc = new Hammer(item[0]);
				
				mc.on("tap", function() {
					setTimeout(function() { // timeout because modal window causes a long press as well
					
						var result = null;
						if (g_config.shots_mode > 0 && string_is_accepted && g_config.use_role_shooter) {
							// for final scores, no need to ask for text
							var result = confirm(S("STRING_FINISH_CONFIRM", string.frame().data.name));
							if (!result) return;
						} else {
							result = prompt(S("STRING_FINISH_PROMPT", string.frame().data.name), string.data.name);
							if (result == null) return;
						}
						
						save_string(string, result);

					}, 250);
				});
				
				/* DMC 4-Dec-2018 "save note" for long-press on SAVE has too much potential for confusion among
				untrained users*/
				// DMC 14-Jan-2019: do "save note" only for RO/ADMIN
				if (ro_check() || admin_check()) {
					mc.on("press", function() {
						var result = prompt("Save string with note? Enter a note.", string.data.note);
						string.crud({ note: result });
					});
				}
				
				panel.append(item);
			}

			// button for submitting scores for acceptance
			if (scores_are_final && !string_is_accepted && !string.data.accepted) {

				panel.append("<hr>");
				var item = toolbar.new_button(panel, "Scorecard<br>Ready!", function() {
					
					var string = cur_string();

					// snapshot list of shooters currently assigned
					var accepted = {};
					if (string.data.role_shooter) {
						var shooter_count = string.data.shooter_count || 1;
						for (var i = 0; i < shooter_count; i++) {
							var id = string.data.role_shooter[i];
							if (id) accepted[id] = false;
						}
					}

					string.crud({ accepted: accepted, who_accepted: {} });
				});
				item.css("font-weight", "bold");
				
			}
			
			// acceptance in progress: show status
			if (string.data.accepted) {

				panel.append("<hr>");
					
				var count = 0;
				for (i in string.data.accepted) {
					if (string.data.accepted[i]) {
						count++;
					}
				}

				if (count < numberof(string.data.accepted)) {
					panel.append(count + " / " + numberof(string.data.accepted) + " Accepted");
					panel.append("<br>");
				}

				// cancel anytime, even if all shooters have accepted
				toolbar.new_button(panel, "Cancel<br>Scorecard", function() {
					string.crud({ accepted: 0, who_accepted: 0 });
				});
				
			}
			
		}
	
		if (string.data.note) {
			panel.append("<br>");
			var item = toolbar.new_button(panel, "Read Note", function() {
				alert(string.data.note);
			});
			item.addClass("button_small");
		}

		// everything between header panel and table
		if (!display_admin_mode()) this.make_string_controls(this.div);
		
		// shots table
		var table = $("<table id=shotstable>");
		
		if (printing) table.css("border", "2px solid");

		var allshots = string.shots();
		var validshots = string.validshots();
		
		if (!display_admin_mode() && shooter_count > 1 && !g_config.basic_mode) {
			this.make_multi_shooter_table(table, validshots);
		} else if (display_admin_mode()) {
			this.make_single_shooter_table(table, allshots);
		} else {
			this.make_single_shooter_table(table, validshots);
		}

		// end shots table
		this.div.append(table);
				
		// scroll to selected shot
		if (this.focus_tr) {
			this.div.scrollTop(this.focus_tr[0].offsetTop - 300);
		}
		
		// hack needed just for for IE		
		set_pos(this.div, "height", $(window).height() - 30);

		//log("shotlist.update() done");
		show(this.div);
	}
	
	this.make_string_controls = function(div) {
		
		var string = cur_string();
		var shooter_count = string.data.shooter_count || 1;
		if (g_config.basic_mode) shooter_count = 1;

		// whether to enable total number of shots display
		var use_shots_mode = 0;
		if (live_mode() && g_config.shots_mode > 0)
			use_shots_mode = g_config.shots_mode;
		if (!live_mode() && string.data.shots_mode > 0)
			use_shots_mode = string.data.shots_mode;
			
		if (g_config.basic_mode) use_shots_mode = 0;

		// whether to display scorekeepers
		var use_role_scorer = (live_mode() && g_config.use_role_scorer) || (!live_mode() && string.data.use_role_scorer);
		var use_role_shooter = (live_mode() && g_config.use_role_shooter) || (!live_mode() && string.data.use_role_shooter);

		if (use_role_scorer && !g_config.basic_mode) {
			
			var color = "#fdd";
			if (string.data.role_scorer) color = "#dfd";

			var div = $("<div>");
			div.css("background-color", color);
			div.css("margin", "3px");
			div.css("padding", "3px");
			div.css("text-align", "center");
			if (user_is_scorer()) div.css("border", "3px solid");

			var scorer_name;
			var item = toolbar.new_text(div, scorer_name = toolbar_military.stringname_Scorer() /*"Scorekeeper" or "Coach"*/);
			item.css("font-size", "14px");
			item.css("font-weight", "bold");

			div.append("<br>");

			var str = "";
			var user_id = string.data.role_scorer || 0;
			if (user_id) str += user_display_name(user_id);
			else str += "None";

			var item = toolbar.new_text(div, str);
			
			// click to assign scorekeeper if it is open
			if (live_mode() && g_config.users_can_assign && !user_id && cur_user() && cur_user().data.name != "admin") {
				div.mousedown(function() {
					if (user_is_shooter() >= 0) return; // prevent if user is assigned shooter
					if (confirm("Assign yourself as the " + scorer_name + "?")) {
						user_assign_scorer(cur_string().id(), cur_user().id());
					}
				});
			}

			this.div.append(div);
		}

		function draw_zero_div(bool_show, div, string, poi, label_text, x_mult, y_mult) {
			if (!bool_show/*undefined or false or true*/ || !poi) return;
            var color = "#dfd";
            var zero_panel = $("<div id='divZeroScope'>");
            zero_panel.css("background-color", color); // color status
            zero_panel.css("margin", "3px");
            zero_panel.css("padding", "3px");
            zero_panel.css("text-align", "center");
            zero_panel.css("border", "3px solid");
			// if (toolbar_military.is_mil()) inner.css("font-size", "24px");

			zero_panel.css("font-size", "24px");

			zero_panel.append("<b>"+label_text+"</b>");
            zero_panel.append('<br>');
            
            var zerocoords = (cur_string().face().data.options &&  cur_string().face().data.options.zerocoords) || {x:0, y:0};
            var x_moa = cur_string().face().mm_to_moa*(poi.x - zerocoords.x);
            var y_moa = cur_string().face().mm_to_moa*(poi.y - zerocoords.y);
            // convert MOA into rounded sight units:
            x_direction=Math.round(x_moa * x_mult);
            y_direction=Math.round(y_moa * y_mult);
            var right_left = x_direction>0? "Left":"Right";
			var up_down = y_direction>0? "Down":"Up";
            zero_panel.append(right_left +" "+ abs(x_direction) +" <br> " + up_down+" "+abs(y_direction));
            div.append(zero_panel);
		}

		function draw_ballistic_truing_div(bool_show, div, string, label_text, x_mult, y_mult) {
			if (!bool_show/*undefined or false or true*/ || !string.data.stats_group_center) return;
	        var cf_config = cur_string().frame().data.config; // current frame config data
	        if (!cf_config || 'object' != typeof cf_config.BallisticTruing) return;

            var color = "#EBF5FB"; // pale blue
            var zero_panel = $("<div>");
            zero_panel.css("background-color", color); // color status
            zero_panel.css("margin", "3px");
            zero_panel.css("padding", "3px");
            zero_panel.css("text-align", "center");
            zero_panel.css("border", "3px solid");
			zero_panel.append("<b>"+label_text+"</b>");
            zero_panel.append('<br>');


            var unit = get_coords_unit();
            var mpier = get_coords_mpier();
            var ndigs = get_coords_ndigs();

            var s_unit = get_speed_unit();
            var s_mpier = get_speed_mpier();
            var s_ndigs = get_speed_ndigs();

			var score = calc_string_score(string.data, cur_shooter); // not just score data but speed stats too

			var sdata = cur_string().data;
			var stats = sdata;
			if (sdata.shooterStats && sdata.shooterStats[cur_shooter]) {
				var stats = sdata.shooterStats[cur_shooter];
			}
			
			if (cf_config.BallisticTruing['Vert-ES']) {
	            zero_panel.append('Vert-ES: '+(stats.stats_height * mpier).toFixed(ndigs) + " " + unit + '<br>');
			}
			if (cf_config.BallisticTruing['Hor-ES']) {
	            zero_panel.append('Hor-ES: '+(stats.stats_width * mpier).toFixed(ndigs) + " " + unit + '<br>');
			}
			if (cf_config.BallisticTruing['VMR']) {
	            zero_panel.append('VMR: '+(score.v_mr * mpier).toFixed(ndigs) + " " + unit + '<br>');
			}
			if (cf_config.BallisticTruing['Velo-Mean']) {
	            zero_panel.append('Velo-Mean: '+(score.v_avg * s_mpier).toFixed(s_ndigs) + " " + s_unit + '<br>');
			}
			if (cf_config.BallisticTruing['Velo-ES']) {
	            zero_panel.append('Velo-ES: '+((score.v_max-score.v_min) * s_mpier).toFixed(s_ndigs) + " " + s_unit + '<br>');
			}
			if (cf_config.BallisticTruing['Velo-StdDev']) {
	            zero_panel.append('Velo-StdDev: '+(score.v_sd * s_mpier).toFixed(s_ndigs) + " " + s_unit + '<br>');
			}
			if (cf_config.BallisticTruing['Velo-StdErr']) {
	            zero_panel.append('Velo-StdErr: '+(score.v_stderr * s_mpier).toFixed(s_ndigs) + " " + s_unit + '<br>');
			}

			// v:
			if (stats && stats.use_string_stats) {
				if (toolbar_military.is_mil()) {
					str += "<hr>ES - " + (stats.stats_group_size * mpier).toFixed(ndigs) + unit;
				} else {
					str += "<hr>Group: " + (stats.stats_group_size * mpier).toFixed(ndigs) + " " + unit;
					str += "<br>(" + (stats.stats_width * mpier).toFixed(ndigs) + "w x ";
					str += (stats.stats_height * mpier).toFixed(ndigs) + "h)";
				}
			}
  
            div.append(zero_panel);
		}

        var f_config = cur_string().frame().data.config; // OPTIONAL - so handle 'undefined'!!
        var moa_to_mils = moa_to_mil(1);


		var zeroadjust_clicksize_moa = mil_to_moa(0.1);
		if (toolbar_military && toolbar_military.ZeroAdjustClicksizes) { // [ {name: "0.1 Mil", size: mil_to_moa(0.1)},... ]
			$.each(toolbar_military.ZeroAdjustClicksizes, function (idx,clicksize_item) {
				if (f_config && f_config.Zero_clicksize == clicksize_item.name) {
					zeroadjust_clicksize_moa = clicksize_item.size;
				}
			});
		}


        draw_zero_div(f_config && f_config.zero_adjust, this.div, string, string.data.stats_group_center, "Zero Adjust", 1/zeroadjust_clicksize_moa, 1/zeroadjust_clicksize_moa);
        draw_zero_div(f_config && f_config.wind_adjust, this.div, string, string.data.stats_last_shot, "Wind Adjust", 1/zeroadjust_clicksize_moa, 1/zeroadjust_clicksize_moa);
        draw_ballistic_truing_div(f_config && f_config.BallisticTruing && f_config.BallisticTruing.enabled,
         this.div, string, "Ballistic Truing", 1/zeroadjust_clicksize_moa, 1/zeroadjust_clicksize_moa);

        draw_zero_div(f_config && f_config.RMRzero, this.div, string, string.data.stats_group_center, "RMR Adjust", 1, 1);
        draw_zero_div(f_config && f_config.BUISzero, this.div, string, string.data.stats_group_center, "BUIS Adjust"
         , moa_to_mils * 10 / 4.8 //  1 click 16" bbl = 4.8cm @ 100m
         , moa_to_mils * 10 / (1.25*4.8/3.5) // 1.25 cm at 100m for 20" bbl, for 16" bbl is 4.8/3.5
        );
        draw_zero_div(f_config && f_config.LSWzero, this.div, string, string.data.stats_group_center, "LSW Adjust", 1, 1);

		// show shooter boxes in all cases except when it would be empty
		if (use_role_shooter || shooter_count > 1 || use_shots_mode || string.face().data.scoringzones_f) {

			// show shooter info boxes
			for (var i = 0; i < shooter_count; i++) {
				
				// whether this is the currently selected shooter
				var selected = (i == cur_shooter);
				
				var score = null; // score data
				if (use_shots_mode) score = calc_string_score(string.data, i);
				
				var color = "#dfd"; // default green
				if (use_role_shooter && (!string.data.role_shooter || !string.data.role_shooter[i])) {
					color = "#fdd"; // no scorer: red
				} else if (use_shots_mode) {
					if (score.count == use_shots_mode) {
						color = "#eee"; // final score: grey
					} else if (score.count > use_shots_mode) {
						color = "#fdd"; // too many shots: red
					}
				}
				
				var div = $("<div>");
				div.css("background-color", color); // color status
				div.css("margin", "3px");
				div.css("padding", "3px");
				div.css("text-align", "center");
				if (selected) div.css("border", "3px solid");

				// assigned shooter cannot view others
				if (!live_mode() || user_is_shooter() < 0) {
					div.mousedown(function() {
						var new_shooter = this.arg;
						if (cur_shooter != new_shooter) {
							cur_shooter = new_shooter;
							update();
						}
					});
					div[0].arg = i;
				}

				// header line
				var str = "";
				if (shooter_count == 2) str = ((i==0)?("L"):("R"));
				else if (shooter_count == 3) str = ((i==0)?("L"):((i==1)?("M"):("R")));
				var header = toolbar.new_text(div, str)
					.css({"font-size": "24px", "font-weight": "bold","padding":"10px"});
				
				var user_id = 0;
				if (use_role_shooter && string.data.role_shooter && string.data.role_shooter[i]) {
					user_id = string.data.role_shooter[i];
				}

				// fclass target setting
				if (!g_config.is_NRA && string.face().data.scoringzones_f) { // target has multiple scoring zones
					var fclass_forced = typeof(user_is_registered(user_id)) == "object";
					var fclass = string.data.fclass_target && string.data.fclass_target[i];
					var text_f = "F-Class", text_sling = "Sling";
					var text_f = "F", text_sling = "Iron";
					var text = fclass ? text_f : text_sling;
					var show_button = !user_id && user_can_affect_shots() && live_mode(); // only if there is no user logged in
					if (!show_button) { // label
						toolbar.new_text(div," (" + text + ")");
					} else { // buttons
						// set F-Class face
						var item = toolbar.new_button(div, text_f, function() {
							var fclass_target = [];
							fclass_target[this.shooter_index] = true;
							string.crud({fclass_target: fclass_target});
							g_socket.emit("recalc_shot_scores", string.id());
						},fclass);
						item[0].shooter_index = i;
						item.addClass("button_small").css("font-size", "14px")//.css("font-weight", "bold")
						// set TR face
						var item = toolbar.new_button(div, text_sling, function() {
							var fclass_target = [];
							fclass_target[this.shooter_index] = false;
							string.crud({fclass_target: fclass_target});
							g_socket.emit("recalc_shot_scores", string.id());
						},!fclass);
						item[0].shooter_index = i;
						item.addClass("button_small").css("font-size", "14px")//.css("font-weight", "bold");
					}
				}
				// DMC bullet diameter, string.bullet_dia gets copied to each shot.bullet_dia
				if (g_config.is_NRA) {
					var f_bullet_dia = cur_string().frame().data.bullet_dia; // OPTIONAL - so handle 'undefined'!!

					var bullet_dia = cur_string().data.bullet_dia;
					var name = "CAL?",idx=-1;
					for (i in shotlist.bullet_dia_list) {
						if (bullet_dia == shotlist.bullet_dia_list[i].dia) {
							name = shotlist.bullet_dia_list[i].name;
							idx = (+i) +1; //  arg, i is a string!
						}
					}
					if (idx == shotlist.bullet_dia_list.length) idx=0;
					if (idx == -1) idx = 0;
					const next_dia = shotlist.bullet_dia_list[idx].dia;
					toolbar.new_button(div, name, function() {
						cur_string().crud({bullet_dia: next_dia});
						g_socket.emit("recalc_shot_scores", string.id());
					}, false);

				}

				
				// assigned shooter name
				if (use_role_shooter) {

					div.append("<br>");
					var item = toolbar.new_text(div,
						toolbar_military.stringname_Shooter(i,shooter_count) // "Firer", "Shooter", "Left Shooter", etc
					).css("font-size", "14px").css("font-weight", "bold")

					div.append("<br>");

					var str = "";
					if (user_id) str += user_display_name(user_id);
					else str += "None";
					var item = toolbar.new_text(div, str);					
				}
				
				if (use_shots_mode) {
					
					// final score
					if (score.count > 0 && score.count <= use_shots_mode) {
						div.append("<br>");
						var scoretext = score.string;
						if (toolbar_military.is_mil()) {
							scoretext = "Score<br>" + scoretext;
						}
						var item = toolbar.new_text(div, scoretext);
		  				item.css("font-size", "24px"); // regular 16px
						if (score.count == use_shots_mode) {
							if (selected) {
								item.css("font-size", "40px");
								if (toolbar_military.is_mil()) item.css("font-size", "24px");
							}
		  					item.css("font-weight", "bold");
		  				}
					}
					
					// score status
					var str = S("STRING_SHOTS", score.count, use_shots_mode); // "3 / 20 shots"
					var bold = false;
					if (score.count == use_shots_mode) {
						str = S("STRING_SHOTS_FINAL");
						bold = true;
					} else if (score.count > use_shots_mode) {
						str = S("STRING_SHOTS_TOOMANY");
						bold = true;
					}
					div.append("<br>");
					var item = toolbar.new_text(div, str);
					if (bold) item.css("font-weight", "bold");
				}
				
				if (score && score.v_avg && !toolbar_military.is_mil()) {
					div.append("<br>");
					toolbar.new_text(div, "v: " + (score.v_avg * 3.281).toFixed(0) + ", sd: " + (score.v_sd * 3.281).toFixed(1));
				}
				
				// click again to assign user as shooter
				// 'selected' : normal shooters forced to look at their own plots can't choose a different slot
				if (live_mode() && use_role_shooter && !g_config.basic_mode && selected && g_config.users_can_assign && !user_id && cur_user() && cur_user().data.name != "admin") {
					div.mousedown(function() {

						// prevent if user is already the scorekeeper
						if (user_is_scorer()) return;
						
						// prevent if user is not registered
						if (!user_is_registered()) {
							alert("You are not registered for this event!");
							return;
						}
						
						var str = toolbar_military.stringname_Shooter(this.arg,shooter_count); // "Firer", "Shooter", "Left Shooter", etc

						if (confirm("Assign yourself as the " + str + "?")) {
							user_assign_shooter(string.id(), cur_user().id(), this.arg);
						}
					});
					div[0].arg = i;
				}
				
				// show button for a shooter to accept score
				if (g_config.option_must_accept && g_config.use_role_shooter && user_id && string.data.accepted && string.data.accepted[user_id] == false) {
					
					// who can see the button: only the shooter himself or [*NOT* scorekeeper] or range officer
					if ((cur_user() && cur_user().id() == user_id) || admin_check() || ro_check() /* || user_is_scorer() */) {
					
						div.append("<br>");
						var item = toolbar.new_button(div, "Sign<br>Scorecard", function() {
							var accepted = {};
							accepted[this.arg] = true;
							var who_accepted = {};
							who_accepted[this.arg] = cur_user().id();
							string.crud({ accepted: accepted, who_accepted: who_accepted });
						});
						item[0].arg = user_id;
						
					// waiting for acceptance
					} else {
						
						div.append("<br>");
						var item = toolbar.new_text(div, "Waiting to accept score...");
						item.css("font-weight", "bold");
						item.css("font-size", "14px");
						
					}
					
				}

				this.div.append(div);
			}
			
		}

		// converting sighter buttons
		// even spectators can convert sighters if not using assigned shooters
		if (!printing && live_mode() && !g_config.basic_mode && (!g_config.use_role_shooter || user_can_affect_shots() || user_is_scorer()) && g_config.sighter_count) {
			
			var shot = cur_string().cur_shot();
			var sighter_selected = shot && shot.shooter_num == cur_shooter && shot.sighter;

			// identify sighters in the first few shots in string
			var sighters = [];
			var shot_count = 0;
			for (var i = 0; i < string.data.shots.length && shot_count < 5; i++) {
				var s2 = string.validshots()[i];
				if (s2.strikeout) continue;
				if (s2.shooter_num != cur_shooter) continue; // removes other shooters and invalid shots
				shot_count++; // my real shots
				if (s2.sighter) sighters.push(i); // sighters only
			}

			// on record button marks that the user has requested to stop sighters
			// will not be set automatically, to be able to revert after striking shots
			if (user_can_affect_shots() && !g_config.sighter_convertible // only for non-convertible mode
				&& (!string.data.on_record || !string.data.on_record[cur_shooter]) // not if shooter is on record
				&& sighters.length == shot_count // all shots in string so far are sighters
				&& (g_config.sighter_count < 0 || sighters.length < g_config.sighter_count)) { // does not exceed limit, if limited

				var item = toolbar.new_button(this.div, S("STRING_GORECORD"), function() {
					result = confirm(S("STRING_GORECORD_CONFIRM"));
					if (result) {
						var rec = [];
						rec[cur_shooter] = true;
						string.crud({on_record: rec});
					}
				});
				item.addClass("wide_button");

			}
			// show "convert" button for sighters on current shooter
			if (user_can_affect_shots()
			 && g_config.sighter_convertible && g_config.sighter_count > 0 // only for limited, convertible sighters
			 && shot_count >= g_config.sighter_count // fired at least the sighter limit
			 && sighters.length <= g_config.sighter_count // not too many sighters
			 && (shot_count == sighters.length || sighter_selected)) { // either selected one of my sighters or all my shots are sighters

				// create a button for each sighter, to convert a subset
				for (var i = sighters.length - 1; i >= 0; i--) { // first button = most recent sighter only

					var cmd = $("<button class='wide_button'>");

					// which shots are included in this conversion button
					var list = [];
					for (var j = i; j <= sighters.length - 1; j++) {
						list.push(sighters[j]);
					}
					cmd[0].arg = list;

					// button text
					var str = S("STRING_CONVERT_ONE");
					if (sighters.length == 2) str = (list.length == 1)?(S("STRING_CONVERT_ONE")):(S("STRING_CONVERT_BOTH"));
					else if (sighters.length == 3) str = (list.length == 1)?(S("STRING_CONVERT_ONE")):((list.length == 2)?(S("STRING_CONVERT_TWO")):(S("STRING_CONVERT_ALL")));
					
					cmd.text(str);

					cmd.mousedown(function() {
						var delta_shots = [];
						for (var j in this.arg) {
							delta_shots[this.arg[j]] = {sighter: false}; // LATER - likely buggy - is arg[j] actually the correct index?
						}
						// string.crud({shots: delta_shots, validshots: delta_validshots});
						string.crud_shots(delta_shots, validshots);
						g_socket.emit("update_string_stats", string.id());
					});

					this.div.append(cmd);
				}

			}

		}
		
	}
	
	this.make_single_shooter_table = function(table, shots) {
		
		var shots_displayed = 0,
            update_again_soon = false;
	
		var speed_unit = g_config.option_speed_unit || "fps";
		var basic_mode = g_config.basic_mode;
		var shot_timelimit = cur_string().frame().data.string_params.shottimer_value || 0;

        table.html('');
		table.append("<tr><th>#</th><th>" + speed_unit + "</th><th style='width: 40%'>score</th></tr>");

		if (cur_string().data.shots_full) {
			table.append($('<tr><td colspan=3 style="font-size:large;font-weight:bold;background-color:red;">---<br>Shotlist FULL<br>---<br>SAVE NOW<br>---</td></tr>')
			);
		}

		var selected = cur_string().cur_shot_index;

		// run through shots and set display number
		var num = 1;
		var abs_num = 1;
		var s_num = 1;
		var i_num = 1; // invalid shots
		var c_num = 1; // classed-as-crossfire shots
		var shots_to_show = [];
		var previous_shot = undefined;
		for (var i in shots) {
			var shot = shots[i];
			if (!shot) continue;
			shot.text = ""; shot.show_shotdisc = true; shot.focus = false; shot.show_in_list = false; shot.idx_internal = i;
			if (display_admin_mode(true)) { // **EVERYTHING** gets displayed
				shot.text = s_num++; shot.show_in_list = true;
			} else {
				var info_prepend = "";
				var info_append = display_info_mode() ? (":"+abs_num++ + ""):"";
				if (!shot.valid) {
					// display_info_mode: show invalid shots, show cross-shots
					// TODO: off-target; single-mic; no-solution
					if (shot.screened_out_theta) { // flagged as a crossfire
						if (display_info_mode()) {
							shot.text = info_prepend+ "C" + c_num++ + info_append; // C1, C2 etc for RO "info" view
							shot.show_in_list = shot.show_shotdisc = true;
						}
					} else if (shot.res && shot.res.xy_valid) {
						if (display_info_mode()) {
							shot.text = info_prepend+"I" + i_num++ + info_append; // all other invalid shots
							shot.show_in_list = shot.show_shotdisc = true;
						}
					} else {
						shot.show_in_list = shot.show_shotdisc = false; // do not display invlaid shots except for above exceptions
					}
				} else { // valid shot; displayed same way for info or ordinary user

					// calculate per-shot time
					if (previous_shot) {
						shot.shottimer = shot.ts - previous_shot.ts;
					}

					if (basic_mode) {
// BROKEN
							shot.text = ""; // don't shown any numbers at all
							var delaytime = g_config.basic_mode_delay;
							if (typeof delaytime == 'number') delaytime *= 1000; // seconds-->milliseconds
							shot.show_in_list = true;
							if (display_info_mode()) {
								;//
							} else {
								if (isShotNearInTimeToCurShot(shot,cur_string(),delaytime)) {
									shot.focus = true;
								} else {
									shot.show_shotdisc = false;
								}
							}
					} else {
						shot.show_in_list = true;
						if (!shot.strikeout) {
							if (shot.sighter) {
								shot.text = info_prepend+"S" + s_num++ + info_append;
							} else {
								shot.text = info_prepend+num++ + info_append;
							}
						}
					}
					//
					if (!shot.strikeout) { // when calculating shot times, ignore struck-out shots
						previous_shot = shot;
					}
				}
			}
			if (shot.show_in_list) shots_to_show.push(shot);
		} // 

		var global_timers = g_config.global_timers || [];
		var events_to_show = [];
		
		for (var i in global_timers) {
			timer = global_timers[i];
			var now = Date.now() - cur_ts_local_error;
			if (timer.starttime == 0) { // timer hasn't started yet
				events_to_show.push({
					ts: now,
					descr:"TIMER:" + pretty_time_fmt(timer.duration/1000)
				});
			} else { // timer running/finished
				var show_time = !g_config.is_NRA || admin_check() || ro_check(); // who can save non-final?

				if (timer.running) { // could be running or finished
					if (timer_expiredQ(timer)) { // finished
						events_to_show.push({ts:timer.endtime,descr:"STOP "+ pretty_time_fmt(timer.duration/1000)/*"FINISHED -" + pretty_time_fmt(timer.duration/1000)*/});
					} else {
		                update_again_soon = true;
						events_to_show.push({ts:timer.endtime,descr:"RUNNING - " + (show_time ? pretty_time_fmt(timer_time_remaining(timer)/1000) : "")});
					}
				} else { // paused
					events_to_show.push({ts:timer.endtime,descr:"PAUSED(" + pretty_time_fmt(timer.time_remaining/1000) + ')'}); // when Paused, use .time_remaining
				}
				events_to_show.push({
					ts:timer.starttime,
					descr:"START " + pretty_time_fmt(timer.duration/1000)
				});
			}

			//console.log(i,global_timers[i]);
			
		}

// timers TODO:
// when PAUSED, show PAUSE line at time that it was paused (so struck-out shots make sense)
		// draw shots in this list in reverse order
		for (var j = shots_to_show.length - 1; j >= 0; j--) {
			var shot = shots_to_show[j], idx_internal = shot.idx_internal;
			shotlist_table_prepend(table,shots_to_show,j,events_to_show);

			shots_displayed++;

			var tr,
                hideByDelay = false;

            if (isShotHiddenByDelay(shot, cur_string())) {
                hideByDelay = true;
                update_again_soon = true;
            }
            
            var tr_mousedown_handler_fn = function(evt) {
				if (user_can_affect_shots()) {
					shotbubble.enabled = !shotbubble.enabled;
					shotbubble.update();
					set_shot_selection_timer();
				}
			};

			if (selected != idx_internal) {
				tr = $("<tr onmousedown=shotlist.user_press_shot(" + idx_internal + ")>");
			} else {
				tr = $("<tr>").addClass("selected");
				tr.mousedown(tr_mousedown_handler_fn);
			}

			if (shot.strikeout) tr.addClass("strikeout");

			var is_late_shot = false;
			shot.show_shottimer = false;
			if (shot_timelimit && 'number' == typeof shot.shottimer) {
				shot.show_shottimer = true;
				if (shot.shottimer > shot_timelimit*1000) {
					is_late_shot = true;
					tr.addClass("late_shot");
				}
			}

			if (shot.invalid_reasons && shot.invalid_reasons.length && shot.invalid_reasons.indexOf("V_ERR_TOOBIG") != -1) {
				tr.addClass("V_ERR_TOOBIG");
			}

			if (display_admin_mode(true) && shot.measured) tr.addClass("measured");

			tr.append("<td>" + shot.text + "</td>");

            if (hideByDelay) {
                tr.append("<td>...</td>");
            } else {
                var v = shot.v;
                if (speed_unit == "fps") v = ms_to_fps(v);
                tr.append("<td>" + round(v) + "</td>");
            }

			var td_score = $("<td class='score'>");
			var value = shot.value;
			if (shot.sighter) value = "(" + shot.value + ")";

			// client-side '*'
			if (cur_string().shot_is_whack(shot)) value += "*";

            if (hideByDelay) {
                value = '...';
            }
            
			td_score.text(value);

			if (selected == idx_internal) shot.focus = true;
			
			if (!printing && shot.focus) {
				tr.addClass("selected top row"); // put borders on 
				td_score.addClass("focus");
				if (selected == idx_internal) {
					this.focus_tr = tr;
				} else {
					tr.addClass("bottom");
				}
			}

			tr.append(td_score);

			table.append(tr);

			if (!printing && selected == idx_internal) {

				var tr = $("<tr class='selected bottom row'>");
				tr.mousedown(tr_mousedown_handler_fn);

				var td = $("<td colspan='3'>").css("white-space", "nowrap").addClass("active-shot-details-td");
                if (hideByDelay) {
                    td.attr('data-hide-by-delay', 'true');
                }
				if (is_late_shot) tr.addClass("late_shot");
				this.update_shot_details(td);

				tr.append(td);
				table.append(tr); // this 'tr' is the additional info for a selected shot

			}

		}
		shotlist_table_prepend(table,shots_to_show,j,events_to_show); // anything after last shot??
		
        if (update_again_soon) {
            // refresh after one second...
            var that = this;
            setTimeout(function() { that.make_single_shooter_table(table, shots); }, 1000);
        }
        
		return shots_displayed;
			
	}

	function shotlist_table_prepend(table,shots_to_show,i,addtl_to_show) {
		// remember we are building this list from newest-shots-first
		// we prepend NEWER things than shot #i (but older than shot i+1)
		// our last call will be with shot == -1
		var ts_before = i >= 0 ? shots_to_show[i].ts : 0;
		var ts_after = i < shots_to_show.length-1 ? shots_to_show[i+1].ts: (new Date("Apr 1 2099") - 0); // bugfix; "Apr 2099" worked on Chrome but not Firefox
		//console.log(i,typeof shots_to_show[i],ts_before,ts_after);
		if ('object' != typeof addtl_to_show) return;
		for (var k in addtl_to_show) {
			var itm = addtl_to_show[k];
			if (itm.ts >= ts_before && itm.ts < ts_after) { // show
				var tr=$('<tr>').addClass("addtl_event");
				if (!itm.html) {
					tr.append($('<td colspan=3>').text(itm.descr));
					//var tr=$('<tr>').addClass("addtl_event").append($('<td colspan=3>').html(itm.descr));
				} else {
					tr.append(itm.html);
					//var tr=$('<tr>').addClass("addtl_event").append(itm.html);
				}
				table.append(tr);
				//console.log(i,tr);
			}
		}
	}

	
	this.make_multi_shooter_table = function(table, shots) {
		var count = cur_string().data.shooter_count;
		var shot_timelimit = cur_string().frame().data.string_params.shottimer_value || 0;
		
		var speed_unit = g_config.option_speed_unit || "fps";

		var tr = $("<tr>");
		var shooter_char;
		if (cur_string().data.shooter_count == 2) shooter_char = ["L", "R"];
		else shooter_char = ["L", "M", "R"];
		for (var i = 0; i < count; i++) {
			tr.append("<th>" + shooter_char[i] + "</th>");
		}
		table.append(tr);

		var selected = cur_string().cur_shot_index;
		
		var shottable = [];
		
		var sighter_num = [0,0,0];
		var record_num = [0,0,0];
		
		var shots_displayed = 0;

		// run through shots and build table from bottom up
		var col = 0;
		var row = 0;
		shottable[0] = [];
		var previous_shot = undefined;
		for (var i = 0; i < shots.length; i++) {
			var shot = shots[i];
			if ('object' == typeof shot) {
				shot.text = ""; shot.show_shotdisc = false; shot.focus = false; shot.show_shottimer = false;
			}
			
			if (!shot || !shot.valid) continue; // skip altogether
			
			if (!(shot.shooter_num >= 0 && shot.shooter_num < count)) continue;
			
			shots_displayed++;
						
			//log("shot " + i + " is valid, assigned to " + shot.shooter_num);
			
			// check per-shot time
			if (previous_shot) {
				shot.shottimer = shot.ts - previous_shot.ts;
				shot.show_shottimer = true;
//console.log("set shottimer",i,shot.shottimer,shots_displayed,shot.valid);
			}
			
			if (shot.shooter_num == count-col-1) {
				if (!shottable[row]) shottable[row] = [];
				shottable[row][count-col-1] = shot;
				//log("inserting at " + row + ", " + col);
				shot.index = i;
				shot.show_shotdisc = true;
				if (shot.strikeout) {
					shot.text = "";
				} else if (shot.sighter) {
					shot.text = "S" + ++sighter_num[shot.shooter_num];
				} else {
					shot.text = ++record_num[shot.shooter_num];
				}
				previous_shot = shot;
			} else {
				i--; // continue and check this shot against next column
			}
			
			col++; // increment to next column
			
			// start next row
			if (col == count) {
				col = 0;
				row++;
			}
		}
		for (var row = shottable.length-1; row >= 0; row--) {
			var tr = $("<tr>");
/*
			var td = $("<td>");
			td.append(row+1);
			td.css("background-color", "#fff");
			td.css("font-size", "14px");
			tr.append(td);
*/
			for (var col = 0; col < count; col++) {
				var td = $("<td>");
				
				var shot = shottable[row][col];
				if (shot) {
					
					var score = $("<span>").addClass("score");
					score.text(shot.value);
					score.css("font-weight", "bold");
					td.append(score);
					
					if (shot.sighter) {
						score.text("(" + shot.value + ")");
					}
					
					if (shot.strikeout) td.addClass("strikeout");
					if (shot_timelimit && shot.show_shottimer) {
						if (shot.shottimer > shot_timelimit*1000) {
							td.addClass("late_shot");
						}
					}
					if (!printing && selected == shot.index) {
						score.addClass("selected"); td.addClass("selected");
						
						td.mousedown(function() {
							if (user_can_affect_shots()) {
								shotbubble.enabled = !shotbubble.enabled;
								shotbubble.update();
								set_shot_selection_timer();
							}
						});
						
						//
						if (shot_timelimit) { // display time taken to break shot
							if (shot.show_shottimer) {
								td.append("<br>" + get_shottimer_string(shot));
							}
						} else {
							var v = shot.v;
							if (speed_unit == "fps") v = ms_to_fps(v);
							td.append("<br>" + round(v));
						}

						
						this.focus_tr = tr;

					} else {
						
						td[0].shot_index = shot.index;
						td.mousedown(function(e) {
							var index = e.currentTarget.shot_index;
							log("selecting shot " + index);
							shotlist.user_press_shot(index);
						});
					}

     			}

     			tr.append(td);
			}
			
			table.append(tr);
		}
		// NOTE: for now, shot delay is BROKEN for pairs/threes

		return shots_displayed;
	}
	
	// draw shot info in table like date, time, position
	this.update_shot_details = function(div) {
		clearTimeout(shotlist.shot_details_timer);

        if (div === undefined) {
            div = $(".active-shot-details-td:first");
        }
		if (div.length === 0) return;

        var hideByDelay = false;
        if (div.attr('data-hide-by-delay') !== undefined) {
            hideByDelay = true;
        }
        
		div.empty();
		
		var shot = cur_string() && cur_string().cur_shot(); // Aubrey error - Sunday, 19 May 2019 7:27 PM
		if (!shot) return;
		
		// date/time
		var date_shot = new Date(shot.ts); // shot time on server
		var date_now = new Date(Date.now() - cur_ts_local_error); // current server time
		var dif_str = "";
		if (date_now < date_shot) {
			dif_str = "";
		} else if (date_now.toDateString() == date_shot.toDateString()) { // today
			var ds = Math.round((date_now.getTime() - date_shot.getTime()) / 1000);
			var s = ds % 60;
			var m = Math.floor(ds / 60) % 60;
			var h = Math.floor(ds / 3600) % 24;
			if (h > 0) dif_str = h + "h " + m + "m ago";
			else dif_str = ((m>0)?(m + "m "):("")) + s + "s ago";
		} else { // "month day"
			var date_words = date_shot.toDateString().split(" ");
			dif_str = date_words[1] + " " + date_words[2];
		}
		
		var shottimer_str = get_shottimer_string(shot);
		if (shottimer_str != '') shottimer_str = ' (' + shottimer_str + ')';

		div.append(date_shot.toLocaleTimeString() + shottimer_str );
		if (dif_str) div.append("<br>(" + dif_str + ")");		

		// display coords for normal users
		if (!display_admin_mode() && g_config.option_shot_xy && 'number' == typeof shot.x && 'number' == typeof shot.y ) {
            var unit = get_coords_unit();
            var mpier = get_coords_mpier();
            var ndigs = get_coords_ndigs();
			var x = shot.x;
			var y = shot.y;
            if (hideByDelay) {
                x = '...';
                y = '...';
            } else {
				x = (x*mpier).toFixed(ndigs);
				y = (y*mpier).toFixed(ndigs);
			}
			div.append("<br>x,y = <span class='boldnumber'>" + x + ", " + y + " " + unit + "</span><br>");
		}

		// theta and phi, for info and admin
		// DMC 22-Mar-2018: shot.res is not always present!! (e.g. a saved shot string from an older,
		// different server version
		if (display_admin_mode(true) || display_info_mode()) {
			div.append("<br>th,ph = <span class='boldnumber'>" +Math.round(shot.res && shot.res.theta*1000)+ "," + Math.round(shot.res && shot.res.phi*1000) + " mil</span>");
		}



		// calibration
		if (display_admin_mode(true) && live_mode()) {
			
			// only for shots with valid solution
			if (typeof(shot.x) == "number" && typeof(shot.y) == "number") {
				
				// display mm position
				var x_mm = Math.round(shot.x);
				var y_mm = Math.round(shot.y);
				div.append("<br>x,y = <span class='boldnumber'>" + x_mm + ", " + y_mm + "</span><br>");
				
				if (!shot.measured) {

					toolbar.new_button(div, S("STRING_CAL_MEASURE"), function(e) {
						e.stopPropagation(); // so the td isnt clicked

						var shot = cur_string().cur_shot();
						var sx = Math.round(shot.x);
						var sy = Math.round(shot.y);
						result = prompt(S("STRING_CAL_MEASURE_PROMPT", sx, sy));
						if (!result) return;
						var cleaned = result.replace( /[^0-9 -.]/g, '').replace( /\s\s+/g, ' ' ).trim();
						var vals = cleaned.split(" ");
						if (vals.length == 2) {
							var mx = Number(vals[0]);
							var my = Number(vals[1]);
							if (!isNaN(mx) && !isNaN(my)) {

								shot.measured = true;
								shot.measured_x = mx;
								shot.measured_y = my;

								cur_string().crud_shot(cur_string().cur_shot_index, {
									measured: true, measured_x: mx, measured_y: my
								});

								update_calibration(); // client side calculation
								return;
							}
						}
						alert(S("STRING_CAL_MEASURE_INVALID", cleaned));
					});

				} else {

					// display measured position
					var mx_mm = Math.round(shot.measured_x);
					var my_mm = Math.round(shot.measured_y);
					div.append("mx,my = <span class='boldnumber'>" + mx_mm + ", " + my_mm + "</span>");

					// remove measurement button
					toolbar.new_button(div, S("STRING_CAL_CANCEL"), function(e) {
						e.stopPropagation(); // so the td isnt clicked

						var shot = cur_string().cur_shot();

						shot.measured = false;

						cur_string().crud_shot(cur_string().cur_shot_index, {
							measured: false
						});

						update_calibration(); // client side calculation
					});

				}
				
			}
			
			// temperature calibration
			div.append("<br><span class='boldnumber'>" + shot.tempC + "</span> C");
			
			function adjust_temp_cal(delta) {
				var string = cur_string();
				var frame = cur_string().frame();
				
				var prev_temp_cal = frame.data.geom.temp_cal;
				var new_temp_cal = prev_temp_cal + delta;
				frame.data.geom.temp_cal = new_temp_cal;

				var crud = new_crud(); // build message to send to server
				crud.frames[frame.id()] = {};
				crud.frames[frame.id()].geom = { temp_cal: new_temp_cal };
				
				crud.strings[string.id()] = {};
				crud.strings[string.id()].shots = [];

				// update all existing shots in string
				var shots = cur_string().shots();
				for (var i in shots) {
					var shot = shots[i];
					if (!shot) continue;

					// new position
					var new_temp = shot.tempC - prev_temp_cal + new_temp_cal;

					// update on server
					crud.strings[string.id()].shots[i] = {
						tempC: new_temp,
						geom: { // for backward compatibility
							temp_cal: new_temp_cal,
						}
					};

				}

				g_socket.emit("crud", crud);
			}
			
			toolbar.new_button(div, "+1", function(e) {
				e.stopPropagation(); // so the td isnt clicked
				adjust_temp_cal(1);
			});

			toolbar.new_button(div, "-1", function(e) {
				e.stopPropagation(); // so the td isnt clicked
				adjust_temp_cal(-1);
			});
			
		}
		
		if (display_admin_mode() || (admin_check() && !live_mode())) {
			// show mics that fired
			var mic_str = "";
			if (shot.mics) { // may not exist with older data
				for (var i = 0; i < shot.mics.length; i++) {
					var t = shot.mics[i];
					if (t == t && t >= 0) {
						mic_str += (i+1);
					}
				}
			}
			div.append("<br>[<span class='boldnumber'>" + mic_str + "</span>]");
		}
		if (toolbar_system.info_plus) {
			$('#infodiv').html(JSON.stringify(shot, null, "    "));
		}
		
		// update once a second for time display
		if (live_mode()) shotlist.shot_details_timer = setTimeout(shotlist.update_shot_details, 1000);
	}

}


function get_coords_unit() {
	var unit = g_config.option_coords || "mm";
	return unit;
}

function get_coords_mpier() {
	var unit = get_coords_unit();
	var mpier = 1; // for m
    if (unit == "inch") {
        mpier = mm_to_inch(1);
    } else if (unit == "moa") {
        mpier = cur_string().face().mm_to_moa;
    } else if (unit == "mil") {
        mpier = moa_to_mil(cur_string().face().mm_to_moa);
    }
	return mpier;
}

function get_coords_ndigs() {
	var unit = get_coords_unit();
	var ndigs = 1; // for mm
    if (unit == "inch") {
        ndigs = 2;
    } else if (unit == "moa") {
        ndigs = 2;
    } else if (unit == "mil") {
        ndigs = 2;
    }
	return ndigs;
}

function get_speed_unit() {
	var unit = g_config.option_speed_unit || "fps";
	return unit;
}

function get_speed_mpier() {
	var unit = get_speed_unit();
	var mpier =  ms_to_fps(1); // default fps
    if (unit == "m/s") {
        mpier = 1;
    }
	return mpier;
}

function get_speed_ndigs() {
	var ndigs = 1;
	return ndigs;
}

function get_shottimer_string(shot) {
	var shottimer_str = '';
	if (shot.show_shottimer) {
		if (shot.shottimer > 999*1000) {
			shottimer_str = '---';
		} else {
			shottimer_str = '' + (shot.shottimer/1000).toFixed(0)+'s';
		}
	}
	return shottimer_str;
}


