
var toolbar_info = new function () {

    this._strikeOutShotPanel = undefined;
    this._refreshInterval = undefined;
    
    // for possible notificationOptions have a look at: https://github.com/akquinet/jquery-toastmessage-plugin/wiki
    this._options = {
        refreshInterval: 5000,      // interval in milliseconds to refresh the content of the panel
        showNotification: true,     // should a popup shown to the user
        notificationOptions: {      // https://github.com/akquinet/jquery-toastmessage-plugin/wiki
            stayTime: 60*1000,        // time in miliseconds before the item has to disappear
            type: 'warning'         // notice, warning, error, success
        }
    };

    this._options10 = {
        // refreshInterval: 5000,      // interval in milliseconds to refresh the content of the panel
        showNotification: true,     // should a popup shown to the user
        notificationOptions: {      // https://github.com/akquinet/jquery-toastmessage-plugin/wiki
            sticky:true,
            position: "middle-left", // top-left, top-center, top-right, middle-left, middle-center, middle-right
            stayTime: 10*1000,        // time in miliseconds before the item has to disappear
            type: 'notice'         // notice, warning, error, success
        }
    };
    
	this.init = function () {
		this.icon = $("<img src='images/info.png' class='toolbar_icon'>");
	};
	
	// whether icon should be shown on main toolbar
    this.ready = function () {
        if (toolbar_military.is_mil()) return false;
        return admin_check() || ro_check();
    };
	
	// try to open this panel (return false to cancel)
	this.show = function () {
		return true;
	};
	
	// try to close this panel (return false to cancel)
	this.hide = function () {
        if (this._refreshInterval !== undefined) {
            clearInterval(this._refreshInterval);
        }
		return true;
	};
	
    this.checkStringForStrikeOutShots = function (sourceString, destinationString) {
        // if we are admin or ro show a pop up if a new strike out shot recognized
        if (
            this._options.showNotification 
            && (admin_check() || ro_check()) 
            && _.isArray(sourceString.shots) 
            && _.some(sourceString.shots, {'strikeout': true})
        ) {
            var message = S('INFO_NEW_STRIKED_OUT_SHOT', destinationString.frame().data.name)
            $().toastmessage('showToast', _.assignIn(this._options.notificationOptions, { text: message }));
        }


        if (
            this._options10.showNotification 
            && g_config.toast_stars !== false
            && (admin_check() || ro_check()) 
            && _.isArray(sourceString.shots) 
            && (
                _.some(sourceString.shots, {'is_starred': true,"valid":true})
             || _.some(sourceString.shots, {'is_starred_ro': true,"valid":true})
            )
        ) {
            var message = 'Starred shot on ' + destinationString.frame().data.name;
            $().toastmessage('showToast', _.assignIn(this._options10.notificationOptions, { text: message }));
        }


    }
    
	// display the panel in the supplied div
	this.draw = function (div) {
		
		var header = toolbar.new_text(div, ""),
            that = this,
            shots = [];

		div.append("<br>");
        
        if (this._refreshInterval !== undefined) {
            clearInterval(this._refreshInterval);
        }
        
        $('<div style="margin-top:15px;">').html(S('INFO_STRIKED_SHOTS_HEADING')).appendTo(div);
        this._strikeOutShotPanel = $('<div />').addClass('strike-out-shots').appendTo(div);
        
        // get all striked out shots from the current strings
        _.forEach(g_strings, function (string) {
            shots = _.concat(shots, _.map(
                _.filter(string.shots(), { 'strikeout': true }),
                function (shot) {
                    // if strikeoutTs isn't present, use the original ts of the shot
                    var ts = shot.strikeoutTs || shot.ts;
                    return {
                    	//frame: shot.frame_name // shot.frame_name is present only for simulated shots
	                    frame : string.frame().data.name
	                    , ts: ts
                    };
                })
            );
        });
        
        // sort the striked out shots by timestamp...
        shots = _.sortBy(shots, 'ts').reverse();

        this._updateStrikeOutShots(shots);
        this._refreshInterval = setInterval(function() {
            that._updateStrikeOutShots(shots);
        }, this._options.refreshInterval);
        
/*
		toolbar.new_button(div, "INSERT MISS", function() {
		alert("not yet implemented");
//			g_socket.emit("lobots_set_green_light", {frame_id:frame.id(),nowvalue:gpio23_state?0:1}); // toggle it
		});
*/
		
	};
    
    this._updateStrikeOutShots = function (shots) {
        if (this._strikeOutShotPanel === undefined) return;
        
        var that = this,
            now = Date.now() - cur_ts_local_error; // current server time (corrected for local device time being different)

        // render a list with stike out shots...
        this._strikeOutShotPanel.html('');
        if (_.some(shots)) {
            _.forEach(shots, function (shot) {
                that._strikeOutShotPanel.append($('<hr>'));
                var $row = $('<div />').addClass('row').appendTo(that._strikeOutShotPanel),
                    time = new Date(now - shot.ts),
                    timeLabel = '';
                if (time.getUTCHours() > 1) {
                    timeLabel = '> 1h';
                } else {
                    timeLabel = ('0' + time.getUTCMinutes()).substr(-2) + ':' + ('0' + time.getSeconds()).substr(-2);
                }
                $row.append($('<span>').html(shot.frame + ':'));
                $row.append($('<span style="margin-left:10px;">').html(timeLabel));
            });
        }
        
    };

};

