/*******************************************************************************
 server_rstat.js: interface to RSTAT program

 $Author: dmc $
 $Revision: 2765 $
 $Date: 2020-06-29 17:03:25 -0300 (Mon, 29 Jun 2020) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/server_rstat.js $
 $Id: server_rstat.js 2765 2020-06-29 20:03:25Z dmc $
*******************************************************************************/

var mysql = undefined;
try {
	mysql = require("mysql");
} catch (e) {
	log("cannot load mysql; RSTAT access not possible", e);
}

var db; // the database connection
var rstatOptions = {
    host     : "192.168.0.15",
    user     : "ODBC",
    password : "xxxx",
    database : "rstat4"
};

function nothing() { };

// one button database sync
// download users
// save strings not yet saved
// display warnings
function rstat_sync(socket) {
	if (!new_storagejs.rstat_ready) {
		console.log("rstat_sync(), data still loading....");
		socket.emit("rstat_log", "error, data still loading...");
		return;
	}
	rstat_connect(socket, function() {
	rstat_load_users(socket, function() {
	rstat_load_events(socket, function() {
	rstat_load_registration(socket, function() {
	rstat_load_matches(socket, function() {
	rstat_load_stages(socket, function() {
	rstat_save_scores(socket, function() {

		// done
		db.end();
		console.log("rstat_sync done!");
		socket.emit("rstat_log", "finished.");
		
		// save changes to strings to disk (like rstat_saved)
		new_storagejs.storage_write_rstat(archive);
		
		broadcast_data();

	});
	});
	});
	});
	});
	});
	});
	
}

// connect to database
function rstat_connect(socket, next) {
	if (!mysql) return;
	db = mysql.createConnection(rstatOptions);
	db.connect(function(err) {
		if (err) { console.log("error connecting RSTAT:", err, "options:", rstatOptions); return; }
		rstat_connected = true;
		console.log("connected to database");
		socket.emit("rstat_log", "connected to database");
		if (next) next();
	});
}

// load/update all user data
function rstat_load_users(socket, next) {
	db.query("select * from shooter", function(err, rows, fields) {
		if (err) { console.log(err); return; }
		console.log("loaded " + rows.length + " users");
		socket.emit("rstat_log", "loaded " + rows.length + " users");

		for (i in rows) {
			var shooter_data = rows[i];
			
			var id = "RSTAT_COMPNO_" + shooter_data.COMPNO;
			
			var user = data.users[id];
			if (!user) data.users[id] = user = {};
			
			user._id = id; // for indexing
			
			// useful information
			user.name = shooter_data.COMPNO;
			user.realname = shooter_data.SNAME;
			user.pref = shooter_data.PREF;
		}
		
		if (next) next();
		
	});
}


function rstat_load_events(socket, next) {
	db.query("SELECT EVNTNO, EventName from eventdesc", function(err, rows, fields) {
		if (err) { console.log(err); return; }
		console.log("loaded " + rows.length + " events");
		socket.emit("rstat_log", "loaded " + rows.length + " events");
		for (i in rows) {
			data.events[rows[i].EVNTNO] = {
				id: rows[i].EVNTNO,
				name: rows[i].EventName,
				matches: {},
				shooters: {}
			};
		}
		if (next) next();
	});
}

function rstat_load_registration(socket, next) {
	db.query("SELECT EVNTNO, COMPNO, GunCat FROM compentry", function(err, rows, fields) {
		if (err) { console.log(err); return; }
		console.log("loaded " + rows.length + " registrations");
		socket.emit("rstat_log", "loaded " + rows.length + " registrations");
		for (i in rows) {
			var event = data.events[rows[i].EVNTNO];
			var user_id = "RSTAT_COMPNO_" + rows[i].COMPNO;
			event.shooters[user_id] = {
				id: user_id,
				guntype: rows[i].GunCat
			};
		}
		if (next) next();
	});
}

function rstat_load_matches(socket, next) {
	db.query("SELECT EVNTNO, MtchNo, DescMtch, Fired FROM matchatb", function(err, rows, fields) {
		if (err) { console.log(err); return; }
		console.log("loaded " + rows.length + " matches");
		socket.emit("rstat_log", "loaded " + rows.length + " matches");
		for (i in rows) {
			if (rows[i].Fired == 0) continue; // skip non-fired matches
			var event = data.events[rows[i].EVNTNO];
			event.matches[rows[i].MtchNo] = {
				id: rows[i].MtchNo,
				name: rows[i].DescMtch,
				stages: {}
			};
		}
		if (next) next();
	});
}

function rstat_load_stages(socket, next) {
	db.query("SELECT EVNTNO, MtchNo, StgNo, YdLine, MaxSc FROM stageatb", function(err, rows, fields) {
		if (err) { console.log(err); return; }
		console.log("loaded " + rows.length + " stages");
		socket.emit("rstat_log", "loaded " + rows.length + " stages");
		for (i in rows) {
			var match = data.events[rows[i].EVNTNO].matches[rows[i].MtchNo];
			if (!match) continue;
			
			var text = "Stage " + rows[i].StgNo + " (" + rows[i].YdLine + ", " + rows[i].MaxSc + ")";

			var dist = 0;
			if (rows[i].YdLine) {
				var list = rows[i].YdLine.replace(/\D/g, ',').split(',');
				for (j in list) {
					var num = Number(list[j]);
					if (num > 20) { dist = num; break; }
				}
			}
			if (dist == 0) {				
				var list = match.name.replace(/\D/g, ',').split(',');
				for (j in list) {
					var num = Number(list[j]);
					if (num > 20) { dist = num; break; }
				}
			}
			
			match.stages[rows[i].StgNo] = {
				id: rows[i].StgNo,
				name: text,
				distance: dist,
				maxscore: rows[i].MaxSc
			};
		}
		if (next) next();
	});
}

// loop through archive and save all strings not yet saved in database
function rstat_save_scores(socket, next) {
	
	var count = 0;
	
	for (var i in archive) {
		var string = archive[i]; // no shots
		
		// already saved
		if (string.rstat_saved) continue;
		
		// mark string has been checked
		string.rstat_saved = true;
		
		// skip if desired
		if (string.rstat_string_action == "abandon") continue;

		// require valid event
		// string should only have event_data if the event is valid
		if (!string.event_data) continue;

		// resolve note and continue
		if (string.rstat_string_action == "resolve") {
			string.note_resolved = true;
		}

		// string has unresolved note
		if (string.note && !string.note_resolved) {
			string.rstat_saved = false;
			socket.emit("rstat_log", "unresolved note", string, -1);
			continue;
		}

		// require shooter roles and shots mode
		if (!string.use_role_shooter || !string.score) continue;
		if (!(string.shots_mode > 0)) continue;

		// shooter count is always set on saved strings
		for (var i = 0; i < string.shooter_count; i++) {

			// requires valid user assigned
			var user_id = 0;
			if (string.role_shooter && string.role_shooter[i]) {
				user_id = string.role_shooter[i];
			} else continue;

			// must exist and be from rstat database
			var user_id_str = "" + user_id;
			if (!data.users[user_id] || user_id_str.substring(0, 13) != "RSTAT_COMPNO_") continue;

			// must be final
			if (string.score[i].count != string.shots_mode) continue;
			
//if (count > 10) continue;
			console.log ("saving score " + string.score[i].string + " for user " + data.users[user_id].realname);
			log ("saving score " + string.score[i].string + " for user " + data.users[user_id].realname);
			
			count++;

			// send to database
			(function(string, i) {
				var action = "insert";

//action = "replace";
				if (string.rstat_string_action == "replace") action = "replace";
				var user_id = string.role_shooter[i];
				var score = string.score[i];
				var query = action + " into scores (EVNTNO, MtchNo, StgNo, COMPNO, SC) values (" + string.event_data.event_id + "," + string.event_data.match_id + "," + string.event_data.stage_id + "," + data.users[user_id].name + "," + score.rstat_score + ")";
				db.query(query, function(err, rows, fields) {
					if (err) {
						string.rstat_saved = false;
						console.log(err);
						socket.emit("rstat_log", err.toString(), string, i);
						return;
					}
				});
			})(string, i);

			// ensure stage zero aggregate exists
			var query = "insert into scores (EVNTNO, MtchNo, StgNo, COMPNO) values (" + string.event_data.event_id + "," + string.event_data.match_id + ",0," + data.users[user_id].name + ")";
			db.query(query, nothing); // ignore the result

		}
		
	}

	socket.emit("rstat_log", "found " + count + " scores to save");
	
	// can't update archive yet: need to wait for db to finish executing queries

	if (next) next();

}



