/*******************************************************************************
 messagehandlerJSON.js: handle single-line JSON messages

 $Author: dmc $
 $Revision: 3218 $
 $Date: 2023-11-16 10:51:01 -0400 (Thu, 16 Nov 2023) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/lib/messagehandlerJSON/cmsg.js $
 $Id: cmsg.js 3218 2023-11-16 14:51:01Z dmc $
*******************************************************************************/

var EventEmitter    = require('events').EventEmitter;
var util            = require('util');
var log             = require('log4js').getLogger('MessageHandlerJSON');

var cmsgHandler   = require('./cmsg.js');
var errHandler   = require('./err.js');

function MessageHandlerFactory() {

    function MessageProxy() {
        this.handlers = {};
        this.numberOfBytesToRead = 0;
        this.rawData = '';
        // this._startConditionCounter = 0;
        this._activeHandler = undefined;
    }

    util.inherits(MessageProxy, EventEmitter);
    
    MessageProxy.prototype.addHandler = function(idString, handler) {
        log.info('Add handler ' + handler.name + ' for id string ' + idString);
        this.handlers[idString] = handler;
    }

    MessageProxy.prototype.addData = function(databytes) {
        if ('string' != typeof databytes) return;
        var lines = databytes.split('\n');
        while (lines && lines.length) {
            this.rawData += lines[0];
            if (lines.length == 1) { // no newlines follow
                return;
            }

            if ("" == this.rawData) {
                // ignore trivial case of nothing-to-processs
            } else {
                // valid JSON?
                var my_JSON = undefined, parse_error = undefined;
                try {
                    my_JSON = JSON.parse(this.rawData)
                } catch (e) {
                    parse_error = e.toString();
                }
                if (!my_JSON) {
                    var err_handler = this.handlers["ERROR"];
                    if (err_handler) err_handler.onMessage(parse_error)
                } else { // it worked
                    // any handlers?
                    for (var i in this.handlers) {
                        // i = "cmsg", undefined, etc
                        var myhandler = this.handlers[i];
                        if (my_JSON[i]) {
                            myhandler.onMessage(my_JSON,i);
                        }
                    }
                }
            }
            this.rawData = '';
            lines.shift();
        }
    }

    this.Proxy = MessageProxy;
    
    // add all available handler...
    this.cmsgHandler  = cmsgHandler;
    this.errHandler  = errHandler;
}

module.exports = new MessageHandlerFactory();